import time
import uuid
import sys
import inspect
import copy
from .containerbase import ContainerBase
from ..namespaces import domTag
from .. import pmtypes
from . import containerproperties as cp


class AbstractStateContainer(ContainerBase):
    NODENAME = domTag('State')

    # these class variables allow easy type-checking. Derived classes will set corresponding values to True
    isSystemContextState = False
    isRealtimeSampleArrayMetricState = False
    isMetricState = False
    isOperationalState = False
    isComponentState = False
    isAlertState = False
    isAlertSignal = False
    isAlertCondition = False
    isMultiState = False
    isContextState = False

    ext_Extension = cp.ExtensionNodeProperty()
    DescriptorVersion = cp.IntegerAttributeProperty('DescriptorVersion', defaultPyValue=0)  # an integer
    StateVersion = cp.IntegerAttributeProperty('StateVersion', defaultPyValue=0)  # an integer
    _props = ('ext_Extension', 'DescriptorVersion', 'StateVersion')

    stateVersion = StateVersion  # lower case for backwards compatibility

    def __init__(self, nsmapper, descriptorContainer, node=None):
        self.descriptorContainer = descriptorContainer
        self.descriptorHandle = descriptorContainer.handle
        super(AbstractStateContainer, self).__init__(nsmapper, node)

        if node is None:
            self.DescriptorVersion = descriptorContainer.DescriptorVersion

    @property
    def nodeName(self):
        return self.NODENAME

    def updateNode(self):
        self.node = self.mkStateNode()

    def mkStateNode(self, tag=None, updateDescriptorVersion=True):
        if updateDescriptorVersion:
            self.updateDescriptorVersion()
        node = super(AbstractStateContainer, self).mkNode(tag, setXsiType=True)
        node.set('DescriptorHandle', self.descriptorHandle)
        return node

    def updateFromNode(self, node):
        """ update self.node with node, and set members.
        Accept node only if descriptorHandle matches"""
        descriptorHandle = node.get('DescriptorHandle')
        if self.descriptorHandle is not None and descriptorHandle != self.descriptorHandle:
            raise RuntimeError(
                'Update from a node with different descriptor handle is not possible! Have "{}", got "{}"'.format(
                    self.descriptorHandle, descriptorHandle))
        super(AbstractStateContainer, self)._updateFromNode(node)
        self.node = node

    def updateFromOtherContainer(self, other, skippedProperties=None):
        if other.__class__ != self.__class__:
            raise RuntimeError('Update from a node with different type is not possible! Have "{}", got "{}"'.format(
                self.__class__.__name__, other.__class__.__name__))
        if other.descriptorHandle != self.descriptorHandle:
            raise RuntimeError(
                'Update from a node with different descriptor handle is not possible! Have "{}", got "{}"'.format(
                    self.descriptorHandle, other.descriptorHandle))

        # update all ContainerProperties
        if skippedProperties is None:
            skippedProperties = []
        self.node = other.node
        for prop_name, _ in self._sortedContainerProperties():
            if prop_name not in skippedProperties:
                new_value = getattr(other, prop_name)
                setattr(self, prop_name, new_value)

    def incrementState(self):
        if self.StateVersion is None:
            self.StateVersion = 1
        else:
            self.StateVersion += 1

    def updateDescriptorVersion(self):
        if self.descriptorContainer is None:
            raise RuntimeError('State {} has no descriptorContainer'.format(self))
        if self.descriptorContainer.DescriptorVersion != self.DescriptorVersion:
            self.DescriptorVersion = self.descriptorContainer.DescriptorVersion

    def __repr__(self):
        return '{} descriptorHandle="{}" StateVersion={}'.format(self.__class__.__name__, self.descriptorHandle,
                                                                 self.StateVersion)


class AbstractOperationStateContainer(AbstractStateContainer):
    NODETYPE = domTag('AbstractOperationState')  # a QName
    isOperationalState = True
    OperatingMode = cp.NodeAttributeProperty('OperatingMode', defaultPyValue=pmtypes.OperatingMode.ENABLED)
    _props = ('OperatingMode',)


class SetValueOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('SetValueOperationState')  # a QName
    AllowedRange = cp.SubElementListProperty([domTag('AllowedRange')], cls=pmtypes.Range)
    _props = ('AllowedRange',)


class SetStringOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('SetStringOperationState')  # a QName
    AllowedValues = cp.SubElementTextListProperty([domTag('AllowedValues'), domTag('Value')])
    _props = ('AllowedValues',)


class ActivateOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('ActivateOperationState')  # a QName


class SetContextStateOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('SetContextStateOperationState')  # a QName


class SetMetricStateOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('SetMetricStateOperationState')  # a QName


class SetComponentStateOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('SetComponentStateOperationState')  # a QName


class SetAlertStateOperationStateContainer(AbstractOperationStateContainer):
    NODETYPE = domTag('SetAlertStateOperationState')  # a QName


class AbstractMetricStateContainer_Base(AbstractStateContainer):
    """
    This class is not in the xml schema hierarchy, it only helps to centrally implement functionality
    """
    isMetricState = True

    @property
    def metricValue(self):
        return self._MetricValue

    @metricValue.setter
    def metricValue(self, metricValueObject):
        if metricValueObject is not None:
            assert isinstance(metricValueObject, self.__class__._MetricValue.valueClass)  # pylint: disable=protected-access
        self._MetricValue = metricValueObject

    def mkMetricValue(self):
        if self._MetricValue is None:
            self._MetricValue = self.__class__._MetricValue.valueClass(self.nsmapper)  # pylint: disable=protected-access
            return self._MetricValue
        else:
            raise RuntimeError('State (handle="{}") already has a metric value'.format(self.handle))


class AbstractMetricStateContainer(AbstractMetricStateContainer_Base):
    BodySite = cp.SubElementListProperty([domTag('BodySite')], cls=pmtypes.CodedValue)
    PhysicalConnector = cp.SubElementProperty([domTag('PhysicalConnector')],
                                              valueClass=pmtypes.PhysicalConnectorInfo)  # optional
    ActivationState = cp.NodeAttributeProperty('ActivationState', impliedPyValue=pmtypes.ComponentActivation.ON)
    ActiveDeterminationPeriod = cp.DurationAttributeProperty('ActiveDeterminationPeriod')  # xsd:duration
    LifeTimePeriod = cp.DurationAttributeProperty('LifeTimePeriod')  # xsd:duration, optional
    _props = ('BodySite', 'PhysicalConnector', 'ActivationState', 'ActiveDeterminationPeriod', 'LifeTimePeriod')


class NumericMetricStateContainer(AbstractMetricStateContainer):
    NODETYPE = domTag('NumericMetricState')
    _MetricValue = cp.SubElementProperty([domTag('MetricValue')], valueClass=pmtypes.NumericMetricValue)
    PhysiologicalRange = cp.SubElementListProperty([domTag('PhysiologicalRange')], cls=pmtypes.Range)
    ActiveAveragingPeriod = cp.DurationAttributeProperty('ActiveAveragingPeriod')  # xsd:duration
    _props = ('_MetricValue', 'PhysiologicalRange', 'ActiveAveragingPeriod')


class StringMetricStateContainer(AbstractMetricStateContainer):
    NODETYPE = domTag('StringMetricState')
    _MetricValue = cp.SubElementProperty([domTag('MetricValue')], valueClass=pmtypes.StringMetricValue)
    _props = ('_MetricValue',)


class EnumStringMetricStateContainer(AbstractMetricStateContainer):
    NODETYPE = domTag('EnumStringMetricState')
    _MetricValue = cp.SubElementProperty([domTag('MetricValue')], valueClass=pmtypes.StringMetricValue)
    _props = ('_MetricValue',)


class RealTimeSampleArrayMetricStateContainer(AbstractMetricStateContainer):
    NODETYPE = domTag('RealTimeSampleArrayMetricState')
    isRealtimeSampleArrayMetricState = True
    _MetricValue = cp.SubElementProperty([domTag('MetricValue')], valueClass=pmtypes.SampleArrayValue)
    PhysiologicalRange = cp.SubElementListProperty([domTag('PhysiologicalRange')], cls=pmtypes.Range)
    _props = ('_MetricValue', 'PhysiologicalRange')
    MetricValue = _MetricValue

    def __repr__(self):
        samplesCount = 0
        if self.metricValue is not None and self.metricValue.Samples is not None:
            samplesCount = len(self.metricValue.Samples)
        return '{} Version={} descriptorHandle="{}" Activation="{}" Samples={}'.format(
            self.__class__.__name__, self.StateVersion, self.descriptorHandle, self.ActivationState, samplesCount)


class DistributionSampleArrayMetricStateContainer(AbstractMetricStateContainer):
    NODETYPE = domTag('DistributionSampleArrayMetricState')
    _MetricValue = cp.SubElementProperty([domTag('MetricValue')], valueClass=pmtypes.SampleArrayValue)
    PhysiologicalRange = cp.SubElementListProperty([domTag('PhysiologicalRange')], cls=pmtypes.Range)
    _props = ('_MetricValue', 'PhysiologicalRange')


class AbstractDeviceComponentStateContainer(AbstractStateContainer):
    isComponentState = True
    CalibrationInfo = cp.SubElementProperty([domTag('CalibrationInfo')], valueClass=pmtypes.CalibrationInfo)
    NextCalibration = cp.SubElementProperty([domTag('NextCalibration')], valueClass=pmtypes.CalibrationInfo)
    PhysicalConnector = cp.SubElementProperty([domTag('PhysicalConnector')],
                                              valueClass=pmtypes.PhysicalConnectorInfo)  # optional

    ActivationState = cp.NodeAttributeProperty('ActivationState')  # pmtypes.ComponentActivation
    OperatingHours = cp.IntegerAttributeProperty('OperatingHours')  # optional, unsigned int
    OperatingCycles = cp.IntegerAttributeProperty('OperatingCycles')  # optional, unsigned int
    _props = (
    'CalibrationInfo', 'NextCalibration', 'PhysicalConnector', 'ActivationState', 'OperatingHours', 'OperatingCycles')


class MdsStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('MdsState')
    OperatingMode = cp.NodeAttributeProperty('OperatingMode',
                                             defaultPyValue=pmtypes.MdsOperatingMode.NORMAL)  # pmtypes.MdsOperatingMode
    Lang = cp.NodeAttributeProperty('Lang', impliedPyValue='en')
    _props = ('OperatingMode', 'Lang')


class ScoStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('ScoState')
    OperationGroup = cp.SubElementListProperty([domTag('OperationGroup')], cls=pmtypes.OperationGroup)
    InvocationRequested = cp.NodeAttributeListProperty('InvocationRequested')  # pm:OperationRef
    InvocationRequired = cp.NodeAttributeListProperty('InvocationRequired')  # pm:OperationRef
    _props = ('OperationGroup', 'InvocationRequested', 'InvocationRequired')


class VmdStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('VmdState')


class ChannelStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('ChannelState')


class ClockStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('ClockState')
    ActiveSyncProtocol = cp.SubElementProperty([domTag('ActiveSyncProtocol')], valueClass=pmtypes.CodedValue)
    ReferenceSource = cp.SubElementListProperty([domTag('ReferenceSource')], cls=pmtypes.ElementWithTextOnly)
    DateAndTime = cp.CurrentTimestampAttributeProperty('DateAndTime')
    RemoteSync = cp.BooleanAttributeProperty('RemoteSync', defaultPyValue=True)
    Accuracy = cp.DecimalAttributeProperty('Accuracy')
    LastSet = cp.TimestampAttributeProperty('LastSet')
    TimeZone = cp.NodeAttributeProperty('TimeZone')  # optional, a time zone string
    CriticalUse = cp.BooleanAttributeProperty('CriticalUse', impliedPyValue=False)  # optional
    _props = ('ActiveSyncProtocol', 'ReferenceSource', 'DateAndTime', 'RemoteSync', 'Accuracy', 'LastSet', 'TimeZone',
              'CriticalUse')

    def diff(self, other):
        """ compares all properties EXCEPT DateAndTime.
        BICEPS says:
        "As the current date/time changes at a high frequency, a change of this value SHALL NOT cause
        an update of the state version unless it has been synchronized either remotely or manually."
        returns a list of strings that describe differences"""
        return super().diff(other, ignore_property_names=['DateAndTime'])


class SystemContextStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('SystemContextState')


class BatteryStateContainer(AbstractDeviceComponentStateContainer):
    NODETYPE = domTag('BatteryState')
    CapacityRemaining = cp.SubElementProperty([domTag('CapacityRemaining')], valueClass=pmtypes.Measurement)  # optional
    Voltage = cp.SubElementProperty([domTag('Voltage')], valueClass=pmtypes.Measurement)  # optional
    Current = cp.SubElementProperty([domTag('Current')], valueClass=pmtypes.Measurement)  # optional
    Temperature = cp.SubElementProperty([domTag('Temperature')], valueClass=pmtypes.Measurement)  # optional
    RemainingBatteryTime = cp.SubElementProperty([domTag('RemainingBatteryTime')],
                                                 valueClass=pmtypes.Measurement)  # optional
    ChargeStatus = cp.NodeAttributeProperty('ChargeStatus')  # Ful, ChB, DisChB, DEB
    ChargeCycles = cp.IntegerAttributeProperty('ChargeCycles')  # Number of charge/discharge cycles.
    _props = ('CapacityRemaining', 'Voltage', 'Current', 'Temperature', 'RemainingBatteryTime', 'ChargeStatus',
              'ChargeCycles')


class AbstractAlertStateContainer(AbstractStateContainer):
    isAlertState = True
    ActivationState = cp.NodeAttributeProperty('ActivationState', defaultPyValue=pmtypes.AlertActivation.ON)
    _props = ('ActivationState',)


class AlertSystemStateContainer(AbstractAlertStateContainer):
    NODETYPE = domTag('AlertSystemState')
    SystemSignalActivation = cp.SubElementListProperty([domTag('SystemSignalActivation')],
                                                       cls=pmtypes.SystemSignalActivation)
    LastSelfCheck = cp.TimestampAttributeProperty('LastSelfCheck')
    SelfCheckCount = cp.IntegerAttributeProperty('SelfCheckCount')
    PresentPhysiologicalAlarmConditions = cp.NodeAttributeListProperty('PresentPhysiologicalAlarmConditions')  # pm:AlertConditionReference, List of HANDLE references
    PresentTechnicalAlarmConditions = cp.NodeAttributeListProperty('PresentTechnicalAlarmConditions')  # pm:AlertConditionReference, List of HANDLE references
    _props = ('SystemSignalActivation', 'LastSelfCheck', 'SelfCheckCount', 'PresentPhysiologicalAlarmConditions',
              'PresentTechnicalAlarmConditions')

    def __repr__(self):
        return '{} descriptorHandle="{}" StateVersion={} LastSelfCheck={} SelfCheckCount={}'.format(
            self.__class__.__name__,
            self.descriptorHandle,
            self.StateVersion,
            self.LastSelfCheck,
            self.SelfCheckCount)


class AlertSignalStateContainer(AbstractAlertStateContainer):
    isAlertSignal = True
    NODETYPE = domTag('AlertSignalState')
    Presence = cp.NodeAttributeProperty('Presence', impliedPyValue=pmtypes.AlertSignalPresence.OFF)
    Location = cp.NodeAttributeProperty('Location', impliedPyValue='Loc')  # 'Loc', 'Rem'
    Slot = cp.IntegerAttributeProperty('Slot')  # unsigned int
    ActualSignalGenerationDelay = cp.DurationAttributeProperty('ActualSignalGenerationDelay')  # xsd:duration
    _props = ('Presence', 'Location', 'Slot', 'ActualSignalGenerationDelay')

    def __init__(self, *args, **kwargs):
        super(AlertSignalStateContainer, self).__init__(*args, **kwargs)
        self.lastUpdated = time.time()

        if self.descriptorContainer.SignalDelegationSupported:
            # Delegable signals should have location Remote according to BICEPS
            self.Location = 'Rem'


class AlertConditionStateContainer(AbstractAlertStateContainer):
    isAlertCondition = True
    NODETYPE = domTag('AlertConditionState')
    ActualConditionGenerationDelay = cp.DurationAttributeProperty('ActualConditionGenerationDelay')  # xsd:duration
    ActualPriority = cp.NodeAttributeProperty('ActualPriority')  # optional, pmtypes.AlertConditionPriority ('Lo', 'Me', 'Hi', 'None')
    Rank = cp.NodeAttributeProperty('Rank', valueConverter=cp.IntegerConverter)  # Integer
    DeterminationTime = cp.TimestampAttributeProperty('DeterminationTime')  # Integer
    Presence = cp.NodeAttributeProperty('Presence', valueConverter=cp.BooleanConverter, impliedPyValue=False)
    _props = ('ActualConditionGenerationDelay', 'ActualPriority', 'Rank', 'DeterminationTime', 'Presence')


class LimitAlertConditionStateContainer(AlertConditionStateContainer):
    NODETYPE = domTag('LimitAlertConditionState')  # a QName
    Limits = cp.SubElementProperty([domTag('Limits')], valueClass=pmtypes.Range,defaultPyValue=pmtypes.Range())  # required, pm:Range
    MonitoredAlertLimits = cp.NodeAttributeProperty('MonitoredAlertLimits',
                                                    defaultPyValue=pmtypes.AlertConditionMonitoredLimits.ALL_OFF)  # required, pm:AlertConditionMonitoredLimits
    AutoLimitActivationState = cp.NodeAttributeProperty('AutoLimitActivationState')  # optional, pm:AlertActivation
    _props = ('Limits', 'MonitoredAlertLimits', 'AutoLimitActivationState')


class AbstractMultiStateContainer(AbstractStateContainer):
    isMultiState = True
    Handle = cp.NodeAttributeProperty('Handle')  # required
    _props = ('Handle',)

    def __init__(self, nsmapper, descriptorContainer, node=None):
        super(AbstractMultiStateContainer, self).__init__(nsmapper, descriptorContainer, node)
        if node is None:
            # auto- generate a handle
            self.Handle = uuid.uuid4().hex

    def updateFromNode(self, node):
        """ update self.node with node, and set members.
        Accept node only if descriptorHandle and Handle match"""
        if self.Handle is not None:  # if self.handle is None, this is an initial init from node, no check for equality.
            handle = node.get('Handle')

            if handle != self.Handle:
                raise RuntimeError(
                    'Update from a node with different handle is not possible! Have "{}", got "{}"'.format(
                        self.Handle, handle))
        super(AbstractMultiStateContainer, self).updateFromNode(node)

    def __repr__(self):
        return '{} descriptorHandle="{}" handle="{}" type={}'.format(self.__class__.__name__, self.descriptorHandle,
                                                                     self.Handle, self.NODETYPE)


class AbstractContextStateContainer(AbstractMultiStateContainer):
    isContextState = True
    Validator = cp.SubElementListProperty([domTag('Validator')], cls=pmtypes.InstanceIdentifier)
    Identification = cp.SubElementListProperty([domTag('Identification')], cls=pmtypes.InstanceIdentifier)
    ContextAssociation = cp.NodeAttributeProperty('ContextAssociation',
                                                  impliedPyValue=pmtypes.ContextAssociation.NO_ASSOCIATION)
    BindingMdibVersion = cp.IntegerAttributeProperty('BindingMdibVersion')
    UnbindingMdibVersion = cp.IntegerAttributeProperty('UnbindingMdibVersion')
    BindingStartTime = cp.TimestampAttributeProperty('BindingStartTime')  # time.time() value (float)
    BindingEndTime = cp.TimestampAttributeProperty('BindingEndTime')  # time.time() value (float)
    _props = ('Validator', 'Identification', 'ContextAssociation', 'BindingMdibVersion', 'UnbindingMdibVersion',
              'BindingStartTime', 'BindingEndTime')


class LocationContextStateContainer(AbstractContextStateContainer):
    NODETYPE = domTag('LocationContextState')
    lc = domTag('LocationDetail')
    PoC = cp.NodeAttributeProperty('PoC', [lc])
    Room = cp.NodeAttributeProperty('Room', [lc])
    Bed = cp.NodeAttributeProperty('Bed', [lc])
    Facility = cp.NodeAttributeProperty('Facility', [lc])
    Building = cp.NodeAttributeProperty('Building', [lc])
    Floor = cp.NodeAttributeProperty('Floor', [lc])
    _props = ('PoC', 'Room', 'Bed', 'Facility', 'Building', 'Floor')

    def updateFromSdcLocation(self, sdc_location):
        self.PoC = sdc_location.poc
        self.Room = sdc_location.rm
        self.Bed = sdc_location.bed
        self.Facility = sdc_location.fac
        self.Building = sdc_location.bld
        self.Floor = sdc_location.flr
        self.ContextAssociation = 'Assoc'

        extensionString = self._mkExtensionstring(sdc_location)
        if not extensionString:
            # schema does not allow extension string of zero length
            extensionString = None
        self.Identification = [pmtypes.InstanceIdentifier(root=sdc_location.root, extensionString=extensionString)]
        self.updateNode()

    def _mkExtensionstring(self, sdcLocation):
        return sdcLocation.mkExtensionStringSdc()

    @classmethod
    def fromSdcLocation(cls, nsmapper, descriptorContainer, handle, sdc_location):
        obj = cls(nsmapper, descriptorContainer)
        obj.Handle = handle
        obj.updateFromSdcLocation(sdc_location)
        return obj


class PatientContextStateContainer(AbstractContextStateContainer):
    NODETYPE = domTag('PatientContextState')
    cd = domTag('CoreData')  # a shortcut
    Givenname = cp.NodeTextProperty([cd, domTag('Givenname')])
    Middlename = cp.NodeTextProperty([cd, domTag('Middlename')])
    Familyname = cp.NodeTextProperty([cd, domTag('Familyname')])
    Birthname = cp.NodeTextProperty([cd, domTag('Birthname')])
    Title = cp.NodeTextProperty([cd, domTag('Title')])
    Sex = cp.NodeTextProperty([cd, domTag('Sex')])
    PatientType = cp.NodeTextProperty([cd, domTag('PatientType')])
    DateOfBirth = cp.DateOfBirthProperty([cd, domTag('DateOfBirth')])
    Height = cp.SubElementProperty([cd, domTag('Height')], valueClass=pmtypes.Measurement)
    Weight = cp.SubElementProperty([cd, domTag('Weight')], valueClass=pmtypes.Measurement)
    Race = cp.SubElementProperty([cd, domTag('Race')], valueClass=pmtypes.CodedValue)
    _props = ('Givenname', 'Middlename', 'Familyname', 'Birthname', 'Title', 'Sex', 'PatientType', 'DateOfBirth',
              'Height', 'Weight', 'Race')

    def setBirthdate(self, dateTimeOfBirth_string):
        """ this method accepts a string, format acc. to XML Schema: xsd:dateTime, xsd:date, xsd:gYearMonth or xsd:gYear
        Internally it holds it as a datetime object, so specific formatting of the dateTimeOfBirth_string will be lost."""
        if not dateTimeOfBirth_string:
            self.DateOfBirth = None
        else:
            datetime = cp.DateOfBirthProperty.mk_value_object(dateTimeOfBirth_string)
            self.DateOfBirth = datetime


class WorkflowContextStateContainer(AbstractContextStateContainer):
    NODETYPE = domTag('WorkflowContextState')
    WorkflowDetail = cp.SubElementProperty([domTag('WorkflowDetail')], valueClass=pmtypes.WorkflowDetail)
    _props = ('WorkflowDetail',)


class OperatorContextStateContainer(AbstractContextStateContainer):
    NODETYPE = domTag('OperatorContextState')
    OperatorDetails = cp.SubElementProperty([domTag('OperatorDetails')], valueClass=pmtypes.BaseDemographics)  # optional
    _props = ('OperatorDetails',)


class MeansContextStateContainer(AbstractContextStateContainer):
    NODETYPE = domTag('MeansContextState')
    # class has no own members


class EnsembleContextStateContainer(AbstractContextStateContainer):
    NODETYPE = domTag('EnsembleContextState')
    # class has no own members


# mapping of states: xsi:type information to classes
# find all classes in this module that have a member "NODETYPE"
classes = inspect.getmembers(sys.modules[__name__],
                             lambda member: inspect.isclass(member) and member.__module__ == __name__)
classes_with_NODETYPE = [c[1] for c in classes if hasattr(c[1], 'NODETYPE') and c[1].NODETYPE is not None]
# make a dictionary from found classes: (Key is NODETYPE, value is the class itself
_state_lookup_by_type = dict([(c.NODETYPE, c) for c in classes_with_NODETYPE])


def getContainerClass(qNameType):
    """
    :param qNameType: a QName instance
    """
    return _state_lookup_by_type.get(qNameType)



from collections import OrderedDict
import urllib

class UrlSchemeError(Exception):
    pass

class SdcLocation(object):
    scheme = 'sdc.ctxt.loc'
    locationDetailRoot = 'sdc.ctxt.loc.detail'
    _url_member_mapping = (('fac', 'fac'), ('bldng', 'bld'), ('flr', 'flr'), ('poc', 'poc'), ('rm', 'rm'), ('bed', 'bed')) # urlName, attrName
    def __init__(self, fac=None, poc=None, bed=None, bld=None,  flr=None, rm=None, root=locationDetailRoot):
        self.root = root
        self.fac = fac   # facility
        self.bld = bld   # building
        self.poc = poc   # point of Care
        self.flr = flr   # floor
        self.rm = rm     # room
        self.bed = bed   # Bed
    

    def mkExtensionStringSdc(self):
        elements = self._getExtensionElementsSdc()
        values = [e[1] for e in elements]
        return '/'.join(values)


    def _getExtensionElementsSdc(self):
        """
        :return: a list of (urlName, value) tuples
        """
        identifiers = []
        for urlName, attrName in self._url_member_mapping:
            value = getattr(self, attrName)
            if value is None:
                value = ''
            identifiers.append((urlName, value))
        return identifiers


    @property
    def scopeStringSdc(self):
        return self._mkScopeString(self._getExtensionElementsSdc())


    def _mkScopeString(self, elements):
        identifiers = []
        queryDict= OrderedDict() # technically an OrderedDict is not necessary here, but I like to keep the Query arguments sorted. Easier testing, simple string compare ;)
        for urlName, value in elements:
            identifiers.append(value)
            if value:
                queryDict[urlName] = value
        identifiers = [urllib.parse.quote(ident) for ident in identifiers]
        slash = urllib.parse.quote('/', safe='')
        loc = slash.join(identifiers)  # this is a bit ugly, but urllib.quote does not touch slashes;
        query = urllib.parse.urlencode(queryDict)
        path = '/{}/{}'.format(urllib.parse.quote(self.root), loc)
        scopeString = urllib.parse.urlunparse(urllib.parse.ParseResult(scheme=self.scheme, netloc=None, path=path, params=None, query=query, fragment=None))
        return scopeString


    def matchingServices(self, services):
        return [s for s in services if self.serviceMatches(s)]
        

    def serviceMatches(self, service):
        return self.anyScopeStringMatches(service.getScopes())


    def anyScopeStringMatches(self, scopes):
        for s in scopes:
            if self.scopeStringMatches(s):
                return True
        return False
    
    
    def scopeStringMatches(self, scope):
        """
        Check if location in scope is inside own location.
        :param scope: url string
        :return: boolean
        """
        try:
            other =  self.__class__.fromScopeString(str(scope))
            return other in self
        except UrlSchemeError:
            # Scope has different scheme, no match
            return False


    def __contains__(self, other):
        if self.root != other.root:
            return False
        for my, notmy in ((self.fac, other.fac),
                          (self.bld, other.bld),
                          (self.poc, other.poc),
                          (self.flr, other.flr),
                          (self.rm, other.rm),
                          (self.bed, other.bed)):
            if my is not None:
                if my != notmy:
                    return False
        return True


    @classmethod
    def fromScopeString(cls, s):
        """
        Construct a Location from a scope string. If url scheme is not 'sdc.ctxt.loc', an UrlSchemeError is raised
        :param s: an url
        :return: a SdcLocation object
        """
        src = urllib.parse.urlsplit(s)
        
        if src.scheme.lower() != cls.scheme:
            raise UrlSchemeError('scheme "{}" not excepted, must be "{}"'.format(src.scheme, cls.scheme))
        dummy, root, path = src.path.split('/')
        root = urllib.parse.unquote(root)
        path = urllib.parse.unquote(path)
        queryDict = dict(urllib.parse.parse_qsl(src.query))
        # make a new argumentsDict with well known keys. This allows to ignore unknown keys that might be present in querydict
        argumentsDict = {}
        for urlName, attrName in cls._url_member_mapping:
            argumentsDict[attrName] = queryDict.get(urlName)
        
        argumentsDict['root'] = root
        return cls(**argumentsDict)
    
    
    def __eq__(self, other):
        attrNames = [e[1] for e in self._url_member_mapping]
        attrNames.append('root')
        try:
            for attrName in attrNames:
                if getattr(self, attrName) !=  getattr(other, attrName):
                    return False
            return True
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other    
    
    def __str__(self):
        return '{} {}'.format(self.__class__.__name__, self.scopeStringSdc)
        


from http.server import BaseHTTPRequestHandler
from .compression import CompressionHandler
from io import BytesIO

class DechunkError(Exception):

    """Raised when could not de-chunk stream.
    """

    pass

class DecompressError(Exception):

    """Raised when could not de-compress stream.
    """

    pass


def mkchunks(body, chunk_size=512):
    """
    convert plain body bytes to chunked bytes
    :param body: bytes
    :param chunk_size: size of chunks
    :return: body converted to chunks ( but still as single bytes array)
    """
    data = BytesIO()
    tail = body
    while True:
        head, tail = tail[:chunk_size], tail[chunk_size:]
        data.write(f'{len(head):x}\r\n'.encode('utf-8'))
        data.write(head)
        data.write(b'\r\n')
        if not head:
            return data.getvalue()


class HTTPReader(CompressionHandler):
    """ Base class that implements decoding of incoming http requests.
    Supported features:
    - read data by content-length
    - handle chunk-encoding
    - handle compression
    """
    @classmethod
    def _read_dechunk(cls, stream):
        """De-chunk HTTP body stream.
        :param file stream: readable file-like object.
        :rtype: bytes
        :raise: DechunkError
        """
        body = []
        CRLF = b'\r\n'
        while True:
            chunk_header = cls._read_until(stream, CRLF)
            chunk_headers = chunk_header.split(b';') # length + optional chunk-extensions (name=value pairs)
            chunk_len, chunk_extensions = chunk_headers[0], chunk_headers[1:] # we do nothing with chunk-extensions...
            if chunk_len is None:
                raise DechunkError(
                    'Could not extract chunk size: unexpected end of data.')

            try:
                chunk_len = int(chunk_len.strip(), 16)
            except (ValueError, TypeError) as err:
                raise DechunkError('Could not parse chunk size: %s' % (err,))

            bytes_to_read = chunk_len
            while bytes_to_read:
                chunk = stream.read(bytes_to_read)
                bytes_to_read -= len(chunk)
                body.append(chunk)

            # chunk ends with \r\n
            crlf = stream.read(2)
            if crlf != CRLF:
                raise DechunkError('No CR+LF at the end of chunk!')
            if chunk_len == 0: # len == 0 indicates end of data
                break
        return b''.join(body)

    @staticmethod
    def _read_until(stream, delimiter, max_bytes=16):
        """Read until we have found the given delimiter.
        :param file stream: readable file-like object.
        :param bytes delimiter: delimiter string.
        :param int max_bytes: maximum bytes to read.
        :rtype: bytes|None
        """

        buf = bytearray()
        delim_len = len(delimiter)

        while len(buf) < max_bytes:
            c = stream.read(1)

            if not c:
                break

            buf += c
            if buf[-delim_len:] == delimiter:
                return bytes(buf[:-delim_len])

    @classmethod
    def read_request_body(cls, http_message, supported_encodings=None):
        """ checks header for content-length, chunk-encoding and compression entries.
        Handles incoming bytes correspondingly.
        @http_message: a http request or response read from network
        :return: bytes
        """
        http_body = None
        cl_string = http_message.headers.get('content-length')
        if cl_string:
            try:
                cl = int(cl_string)
                http_body = http_message.rfile.read(cl)
            except TypeError:
                http_body = http_message.rfile.read()
        if http_body is None:
            transfer_encoding = http_message.headers.get('transfer-encoding')
            if transfer_encoding is not None and transfer_encoding.lower() == 'chunked':
                http_body = cls._read_dechunk(http_message.rfile)
        if http_body is None:
            http_body = http_message.rfile.read()

        # if we get compressed content then we check against server setting
        # if it matches continue and decompress
        # if current server setting is any, use whatever client has provided in content-encoding header
        actual_enc = http_message.headers.get('content-encoding')
        if actual_enc:
            supported_encs = supported_encodings or cls.available_encodings
            if actual_enc in supported_encs:
                http_body = cls.decompress(http_body, actual_enc)
            else:
                raise DecompressError('content-encoding "{}" is not supported',actual_enc )
        return http_body

    @classmethod
    def read_response_body(cls, http_response, supported_encodings=None):
        """ checks header for content-length, chunk-encoding and compression entries.
        Handles incoming bytes correspondingly.
        @http_message: a http request or response read from network
        :return: bytes
        """
        http_body = None
        cl_string = http_response.getheader('content-length')
        if cl_string:
            try:
                cl = int(cl_string)
                http_body = http_response.read(cl)
            except TypeError:
                http_body = http_response.read()
        if http_body is None:
            transfer_encoding = http_response.getheader('transfer-encoding')
            if transfer_encoding is not None and transfer_encoding.lower() == 'chunked':
                # de-chunking is done by http client. we just need to read until no more data available
                http_body = []
                tmp = http_response.read()
                while tmp:
                    http_body.append(tmp)
                    tmp = http_response.read()
                http_body = b''.join(http_body)
        if http_body is None:
            http_body = http_response.read()

        # if we get compressed content then we check against server setting
        # if it matches continue and decompress
        # if current server setting is any, use whatever client has provided in content-encoding header
        actual_enc = http_response.getheader('content-encoding')
        if actual_enc:
            supported_encs = supported_encodings or cls.available_encodings
            if actual_enc in supported_encs:
                http_body = cls.decompress(http_body, actual_enc)
            else:
                raise DecompressError('content-encoding "{}" is not supported',actual_enc )
        return http_body


class HTTPRequestHandler(BaseHTTPRequestHandler, CompressionHandler):
    """ Base class that implements decoding of incoming http requests.
    Supported features:
    - read data by content-length
    - handle chunk-encoding
    - handle compression
    """
    protocol_version = "HTTP/1.1"  # this enables keep-alive

    def _read_request(self):
        """ checks header for content-length, chunk-encoding and compression entries.
        Handles incoming bytes correspondingly.
        :return: http body as bytes
        """
        return HTTPReader.read_request_body(self)

    def _compressIfRequired(self, response_bytes):
        """Compress response if header of request indicates that other side
        accepts one of our supported compression encodings"""
        accepted_enc = CompressionHandler.parseHeader(self.headers.get('accept-encoding'))
        for enc in accepted_enc:
            if enc in self.server.supportedEncodings:
                response_bytes = self.compressPayload(enc, response_bytes)
                self.send_header('Content-Encoding', enc)
                break
        return response_bytes

    def log_request(self, *args, **kwargs):
        pass   # supress printing of every request to stderr



import os
import traceback
from urllib import parse

from lxml import etree as etree_

from . import loghelper
from .namespaces import Prefix_Namespace as Prefixes


def mk_schema_validator(schema_resolver: etree_.Resolver) -> etree_.XMLSchema:
    parser = etree_.XMLParser(resolve_entities=True)
    parser.resolvers.add(schema_resolver)
    # create a schema that includes all used schemas into a single one
    all_included = f"""<?xml version="1.0" encoding="UTF-8"?>
    <xsd:schema xmlns:xsd="http://www.w3.org/2001/XMLSchema" elementFormDefault="qualified">
     <xsd:import namespace="http://www.w3.org/2003/05/soap-envelope" schemaLocation="http://www.w3.org/2003/05/soap-envelope"/>
     <xsd:import namespace="http://schemas.xmlsoap.org/ws/2004/08/eventing" schemaLocation="http://schemas.xmlsoap.org/ws/2004/08/eventing"/>
     <xsd:import namespace="http://schemas.xmlsoap.org/ws/2004/09/mex" schemaLocation="http://schemas.xmlsoap.org/ws/2004/09/mex"/>
     <xsd:import namespace="http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01" schemaLocation="http://docs.oasis-open.org/ws-dd/discovery/1.1/os/wsdd-discovery-1.1-schema-os.xsd"/>
     <xsd:import namespace="http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01" schemaLocation="http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01"/>
     <xsd:import namespace="http://www.w3.org/2005/08/addressing" schemaLocation="http://www.w3.org/2006/03/addressing/ws-addr.xsd"/>
     <xsd:import namespace="http://schemas.xmlsoap.org/wsdl/" schemaLocation="http://schemas.xmlsoap.org/wsdl/"/>
     <xsd:import namespace="{Prefixes.MSG.namespace}" schemaLocation="http://standards.ieee.org/downloads/11073/11073-10207-2017/BICEPS_MessageModel.xsd"/>
     </xsd:schema>""".encode('utf-8')

    elem_tree = etree_.fromstring(all_included, parser=parser, base_url='C://')
    return etree_.XMLSchema(etree=elem_tree)


def _needs_normalize(filename):
    return filename.endswith('ExtensionPoint.xsd') or \
           filename.endswith('BICEPS_ParticipantModel.xsd') or \
           filename.endswith('BICEPS_MessageModel.xsd')


class SchemaResolver(etree_.Resolver):

    def __init__(self, base_definitions, log_prefix=None):
        super().__init__()
        self._base_definitions = base_definitions
        self._logger = loghelper.getLoggerAdapter('sdc.schema_resolver', log_prefix)

    def resolve(self, url, id, context):  # pylint: disable=unused-argument, redefined-builtin, invalid-name
        try:
            return self._resolve(url, id, context)
        except:
            self._logger.error(traceback.format_exc())

    def _resolve(self, url, id, context):  # pylint: disable=unused-argument, redefined-builtin, invalid-name
        # first check if there is a lookup defined
        self._logger.debug('try to resolve {}', url)
        path = self._base_definitions.get_schema_file_path(url)
        if path:
            self._logger.debug('could resolve url {} via lookup to {}', url, path)
        else:
            # no lookup, parse url
            parsed = parse.urlparse(url)
            if parsed.scheme == 'file':
                path = parsed.path  # get the path part
            else:  # the url is a path
                path = url
            if path.startswith('/') and path[2] == ':':  # invalid construct like /C:/Temp
                path = path[1:]
            path = parse.unquote(path)  # url decode path

        if not os.path.exists(path):
            self._logger.error('no schema file for url "{}": resolved to "{}", but file does not exist', url, path)
            return None
        with open(path, 'rb') as my_file:
            xml_text = my_file.read()
        if _needs_normalize(path):
            xml_text = self._base_definitions.normalizeXMLText(xml_text)
        return self.resolve_string(xml_text, context, base_url=path)




import weakref
from lxml import etree as etree_
import urllib

from sdc11073 import namespaces
from .. import loghelper
from ..namespaces import msgTag, domTag, QN_TYPE, nsmap, DocNamespaceHelper
from ..namespaces import Prefix_Namespace as Prefix
from ..pysoap.soapenvelope import Soap12Envelope, WsAddress, GenericNode, ExtendedDocumentInvalid
from ..safety import SafetyInfoHeader

class HostedServiceClient(object):
    """ Base class of clients that call hosted services of a dpws device."""
    VALIDATE_MEX = False # workaraound as long as validation error due to missing dpws schema is not solved
    subscribeable_actions = tuple()
    #def __init__(self, soapClient, dpws_hosted, porttype, validate, sdc_definitions, bicepsParser, log_prefix=''):
    def __init__(self, soapClient, dpws_hosted, porttype, sdc_definitions, log_prefix=''):
        """
        :param simple_xml_hosted_node: a "Hosted" node in a simplexml document
        """
        self.endpoint_reference = dpws_hosted.endpointReferences[0]
        self._url = urllib.parse.urlparse(self.endpoint_reference.address)
        self.porttype = porttype
        self._logger = loghelper.getLoggerAdapter('sdc.client.{}'.format(porttype), log_prefix)
        self._operationsManager = None
        #self._validate = validate
        self._sdc_definitions = sdc_definitions
        #self._bicepsParser = bicepsParser
        self.soapClient = soapClient
        self.log_prefix = log_prefix
        self._mdib_wref = None
        self.predefined_actions = {} # calculated actions for subscriptions

        for s in self.subscribeable_actions:
            self.predefined_actions[s] = self._getActionString(s)

    def register_mdib(self, mdib):
        """ Client sometimes must know the mdib data (e.g. Set service, activate method)."""
        if mdib is not None and self._mdib_wref is not None:
            raise RuntimeError('Client "{}" has already an registered mdib'.format(self.porttype))
        self._mdib_wref = None if mdib is None else weakref.ref(mdib)


    def setOperationsManager(self, operationsManager):
        self._operationsManager = operationsManager


    def _callOperation(self, soapEnvelope, request_manipulator=None):
        return self._operationsManager.callOperation(self, soapEnvelope, request_manipulator)


    def getSubscribableActions(self):
        """ action strings only predefined"""
        return self.predefined_actions.values()

    def _getActionString(self, methodName):
        actions_lookup = self._sdc_definitions.Actions
        try:
            return getattr(actions_lookup, methodName)
        except AttributeError: # fallback, if a definition is missing
            return '{}/{}/{}'.format(self._sdc_definitions.ActionsNamespace, self.porttype, methodName)

    def __repr__(self):
        return '{} "{}" endpoint = {}'.format(self.__class__.__name__, self.porttype, self.endpoint_reference)


    def postSoapEnvelope(self, soapEnvelope, msg, request_manipulator=None):
        return self.soapClient.postSoapEnvelopeTo(self._url.path, soapEnvelope, msg=msg, request_manipulator=request_manipulator)

    def _mkSetMethodSoapEnvelope(self, methodName, operationHandle, requestNodes, additionalNamespaces=None):
        """ helper to create the soap envelope
        :param methodName: last element of name of the called action
        :param operationHandle: handle name as string
        :param requestNodes: a list of etree_ nodes that will become Subelement of Method name element
        """
        soapBodyNode = etree_.Element( msgTag(methodName))
        ref = etree_.SubElement(soapBodyNode, msgTag('OperationHandleRef'), attrib={QN_TYPE: '{}:HandleRef'.format(Prefix.PM.prefix)}, nsmap=Prefix.partialMap(Prefix.PM))
        ref.text = operationHandle
        for n in requestNodes:
            soapBodyNode.append(n)
        if additionalNamespaces:
            my_ns = Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG, *additionalNamespaces)
        else:
            my_ns = Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG)

        sih = self._mkOptionalSafetyHeader(soapBodyNode, operationHandle) # a header or None

        soapEnvelope = Soap12Envelope(my_ns)
        action = self._getActionString(methodName)
        soapEnvelope.setAddress(WsAddress(action=action, to=self.endpoint_reference.address))
        if sih is not None:
            soapEnvelope.addHeaderObject(sih)

        soapEnvelope.addBodyElement(soapBodyNode)
        return soapEnvelope


    def _mkGetMethodEnvelope(self, method, params = None):
        action = self._getActionString(method)
        bodyNode = etree_.Element(msgTag(method))
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.WSA, Prefix.MSG))
        soapEnvelope.setAddress(WsAddress(action=action,
                                          to=self.endpoint_reference.address))
        if params:
            for p in params:
                bodyNode.append(p)
        soapEnvelope.addBodyObject(GenericNode(bodyNode))

        return soapEnvelope

    def _callGetMethod(self, method, params = None, request_manipulator=None):
        self._logger.info('calling {} on {}:{}', method, self._url.netloc, self._url.path)
        soapEnvelope = self._mkGetMethodEnvelope(method, params)
        returnedEnvelope = self.postSoapEnvelope(soapEnvelope, msg='get {}'.format(method),
                                                 request_manipulator=request_manipulator)
        return returnedEnvelope

    def _mkSoapEnvelope(self, methodName, xmlBodyString=None, additionalHeaders=None):
        action = self._getActionString(methodName)
        soapEnvelope = Soap12Envelope(Prefix.partialMap(Prefix.S12, Prefix.MSG, Prefix.WSA))
        soapEnvelope.setAddress(WsAddress(action=action, to=self.endpoint_reference.address))
        if additionalHeaders is not None:
            for h in additionalHeaders:
                soapEnvelope.addHeaderObject(h)
        if xmlBodyString is not None:
            soapEnvelope.addBodyString(xmlBodyString)
        return soapEnvelope


    def _mkSoapEnvelopeWithEtreeBody(self, methodName, etreeBody=None, additionalHeaders=None):
        tmp = etree_.tostring(etreeBody)
        return self._mkSoapEnvelope(methodName, tmp, additionalHeaders)


    def _callMethodWithXMLStringArgument(self, portTypeName, methodName, xmlStringArgument=None, additionalHeaders=None):
        soapEnvelope = self._mkSoapEnvelope(methodName, xmlStringArgument, additionalHeaders)
        retEnvelope = self.postSoapEnvelope(soapEnvelope, msg='port {} method {}'.format(portTypeName, methodName))
        return retEnvelope


    def _callMethodWithEtreeNodeArgument(self, portTypeName, methodName, etreeNodeArgument=None, additionalHeaders=None):
        tmp = etree_.tostring(etreeNodeArgument)
        return self._callMethodWithXMLStringArgument(portTypeName, methodName, tmp, additionalHeaders)


    def _mkOptionalSafetyHeader(self, soapBodyNode, operationHandle):

        if self._mdib_wref is not None:
            op_descriptor = self._mdib_wref().descriptions.handle.getOne(operationHandle, allowNone=True)
            if op_descriptor is not None and op_descriptor.SafetyReq is not None:
                mdib_node, mdib_version_group = self._mdib_wref().reconstructMdibWithContextStates()
                return self._mkSoapSafetyHeader(soapBodyNode, op_descriptor.SafetyReq, mdib_node)
        return None


    def _mkSoapSafetyHeader(self, soapBodyNode, t_SafetyReq, mdibNode):
        dualChannelSelectors = {}
        safetyContextSelectors = {}

        if not t_SafetyReq.DualChannelDef:
            self._logger.info('no DualChannel selectors specified')
        else:
            for sel in  t_SafetyReq.DualChannelDef.Selector:
                selectorId = sel.Id
                selectorPath = sel.text
                values = soapBodyNode.xpath(selectorPath, namespaces=mdibNode.nsmap)
                if len(values) == 1:
                    self._logger.debug('DualChannel selector "{}": value = "{}", path= "{}"', selectorId, values[0], selectorPath)
                    dualChannelSelectors[selectorId] = str(values[0]).strip()
                elif len(values) == 0:
                    self._logger.error('DualChannel selector "{}": no value found! path= "{}"', selectorId, selectorPath)
                else:
                    self._logger.error('DualChannel selector "{}": path= "{}", multiple values found: {}', selectorId, selectorPath, values)

        if not t_SafetyReq.SafetyContextDef:
            self._logger.info('no Safety selectors specified')
        else:
            for sel in  t_SafetyReq.SafetyContextDef.Selector:
                selectorId = sel.Id
                selectorPath = sel.text
                # check the selector, there is a potential problem with the starting point of the xpath search path:
                if selectorPath.startswith('//'):
                    # double slashes means that the matching pattern can be located anywhere in the dom tree.
                    # No problem.
                    pass #
                elif selectorPath.startswith('/'):
                    # Problem! if the selector starts with a single slash, this is a xpath search that starts at the document root.
                    # But the convention is that the xpath search shall start from the top level element (=> without the toplevel element in the path)
                    # In order to follow this convention, remove the leading slash and start the search relative to the lop level node.
                    selectorPath = selectorPath[1:]
                values =  mdibNode.xpath(selectorPath, namespaces=mdibNode.nsmap)
                if len(values) == 1:
                    self._logger.debug('Safety selector "{}": value = "{}"  path= "{}"', selectorId, values[0], selectorPath)
                    safetyContextSelectors[selectorId] = str(values[0]).strip()
                elif len(values) == 0:
                    self._logger.error('Safety selector "{}":  no value found! path= "{}"', selectorId, selectorPath)
                else:
                    self._logger.error('Safety selector "{}": path= "{}", multiple values found: {}', selectorId, selectorPath, values)

        if dualChannelSelectors or safetyContextSelectors:
            return SafetyInfoHeader(dualChannelSelectors, safetyContextSelectors)
        else:
            return None


class GetServiceClient(HostedServiceClient):

    def getMdDescriptionNode(self, requestedHandles=None, request_manipulator=None):
        """
        :param requestedHandles: None if all descriptors shall be requested, otherwise a list of handles
        """
        requestparams = []
        if requestedHandles is not None:
            for h in requestedHandles:
                node = etree_.Element(msgTag('HandleRef'))
                node.text = h
                requestparams.append(node)
        resultSoapEnvelope = self._callGetMethod('GetMdDescription', params=requestparams,
                                                 request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode

    def getMdib(self, request_manipulator=None):
        resultSoapEnvelope = self._callGetMethod('GetMdib', request_manipulator=request_manipulator)
        return resultSoapEnvelope

    def getMdibNode(self, request_manipulator=None):
        resultSoapEnvelope = self._callGetMethod('GetMdib', request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode

    def getMdState(self, requestedHandles=None, request_manipulator=None):
        """
        :param requestedHandles: None if all states shall be requested, otherwise a list of handles
        """
        requestparams = []
        if requestedHandles is not None:
            for h in requestedHandles:
                node = etree_.Element(msgTag('HandleRef'))
                node.text = h
            requestparams.append(node)

        resultSoapEnvelope = self._callGetMethod('GetMdState', params=requestparams,
                                                 request_manipulator=request_manipulator)
        return resultSoapEnvelope

    def getMdStateNode(self, requestedHandles=None, request_manipulator=None):
        """
        :param requestedHandles: None if all states shall be requested, otherwise a list of handles
        """
        return self.getMdState(requestedHandles, request_manipulator=request_manipulator).msgNode


class SetServiceClient(HostedServiceClient):
    subscribeable_actions = ('OperationInvokedReport',)

    def setNumericValue(self, operationHandle, requestedNumericValue, request_manipulator=None):
        """ call SetNumericValue Method of device
        :param operationHandle: a string
        :param requestedNumericValue: int or float or a string representing a decimal number
        @return a Future object
        """
        self._logger.info('setNumericValue operationHandle={} requestedNumericValue={}',
                          operationHandle, requestedNumericValue)
        soapEnvelope = self._mkRequestedNumericValueEnvelope(operationHandle, requestedNumericValue)
        return self._callOperation(soapEnvelope, request_manipulator=request_manipulator)

    def setString(self, operationHandle, requestedString, request_manipulator=None):
        """ call SetString Method of device
        :param operationHandle: a string
        :param requestedString: a string
        @return a Future object
        """
        self._logger.info('setString operationHandle={} requestedString={}',
                          operationHandle, requestedString)
        soapEnvelope = self._mkRequestedStringEnvelope(operationHandle, requestedString)
        return self._callOperation(soapEnvelope, request_manipulator=request_manipulator)

    def setAlertState(self, operationHandle, proposedAlertState, request_manipulator=None):
        """The SetAlertState method corresponds to the SetAlertStateOperation objects in the MDIB and allows the modification of an alert.
        It can handle a single proposed AlertState as argument (only for backwards compatibility) and a list of them.
        :param operationHandle: handle name as string
        :param proposedAlertState: domainmodel.AbstractAlertState instance or a list of them
        """
        self._logger.info('setAlertState operationHandle={} requestedAlertState={}',
                          operationHandle, proposedAlertState)
        if hasattr(proposedAlertState, 'NODETYPE'):
            # this is a state container. make it a list
            proposedAlertState = [proposedAlertState]
        soapEnvelope = self._mkSetAlertEnvelope(operationHandle, proposedAlertState)
        return self._callOperation(soapEnvelope, request_manipulator=request_manipulator)

    def setMetricState(self, operationHandle, proposedMetricStates, request_manipulator=None):
        """The SetMetricState method corresponds to the SetMetricStateOperation objects in the MDIB and allows the modification of metric states.
        :param operationHandle: handle name as string
        :param proposedMetricStates: a list of domainmodel.AbstractMetricState instance or derived class
        """
        self._logger.info('setMetricState operationHandle={} requestedMetricState={}',
                          operationHandle, proposedMetricStates)
        soapEnvelope = self._mkSetMetricStateEnvelope(operationHandle, proposedMetricStates)
        return self._callOperation(soapEnvelope, request_manipulator=request_manipulator)

    def activate(self, operationHandle, value, request_manipulator=None):
        """ an activate call does not return the result of the operation directly. Instead you get an transaction id,
        and will receive the status of this transaction as notification ("OperationInvokedReport").
        This method returns a "future" object. The future object has a result as soon as a final transaction state is received.
        :param operationHandle: a string
        :param value: a string
        @return: a concurrent.futures.Future object
        """
        # make message body
        self._logger.info('activate handle={} value={}', operationHandle, value)
        soapBodyNode = etree_.Element(msgTag('Activate'), attrib=None, nsmap=nsmap)
        ref = etree_.SubElement(soapBodyNode, msgTag('OperationHandleRef'))
        ref.text = operationHandle
        argNode = None
        if value is not None:
            argNode = etree_.SubElement(soapBodyNode, msgTag('Argument'))
            argVal = etree_.SubElement(argNode, msgTag('ArgValue'))
            argVal.text = value

        # look for safety context in mdib
        sih = self._mkOptionalSafetyHeader(soapBodyNode, operationHandle)
        if sih is not None:
            sih = [sih]

        soapEnvelope = self._mkSoapEnvelopeWithEtreeBody('Activate', soapBodyNode, additionalHeaders=sih)
        futureObject = self._callOperation(soapEnvelope, request_manipulator=request_manipulator)
        return futureObject

    def setComponentState(self, operationHandle, proposedComponentStates, request_manipulator=None):
        """
        The setComponentState method corresponds to the SetComponentStateOperation objects in the MDIB and allows to insert or modify context states.
        :param operationHandle: handle name as string
        :param proposedComponentStates: a list of domainmodel.AbstractDeviceComponentState instances or derived class
        :return: a concurrent.futures.Future
        """
        tmp = ', '.join(['{}(descriptorHandle={})'.format(st.__class__.__name__, st.descriptorHandle)
                         for st in proposedComponentStates])
        self._logger.info('setComponentState {}', tmp)
        soapEnvelope = self._mkSetComponentStateEnvelope(operationHandle, proposedComponentStates)
        self._logger.debug('setComponentState sends {}', lambda: soapEnvelope.as_xml(pretty=True))
        futureObject = self._callOperation(soapEnvelope, request_manipulator=request_manipulator)
        return futureObject

    def _mkRequestedNumericValueEnvelope(self, operationHandle, requestedNumericValue):
        """create soap envelope, but do not send it. Used for unit testing"""
        requestedValueNode = etree_.Element(msgTag('RequestedNumericValue'),
                                            attrib={QN_TYPE: '{}:decimal'.format(Prefix.XSD.prefix)})
        requestedValueNode.text = str(requestedNumericValue)
        return self._mkSetMethodSoapEnvelope('SetValue', operationHandle, [requestedValueNode],
                                             additionalNamespaces=[Prefix.XSD])

    def _mkRequestedStringEnvelope(self, operationHandle, requestedString):
        """create soap envelope, but do not send it. Used for unit testing"""
        requestedStringNode = etree_.Element(msgTag('RequestedStringValue'),
                                             attrib={QN_TYPE: '{}:string'.format(Prefix.XSD.prefix)})
        requestedStringNode.text = requestedString
        return self._mkSetMethodSoapEnvelope('SetString', operationHandle, [requestedStringNode],
                                             additionalNamespaces=[Prefix.XSD])

    def _mkSetAlertEnvelope(self, operationHandle, proposedAlertStates):
        """create soap envelope, but do not send it. Used for unit testing
        :param proposedAlertStates: a list AbstractAlertStateContainer or derived class """
        _proposedAlertStates = [p.mkCopy() for p in proposedAlertStates]
        for p in _proposedAlertStates:
            p.nsmapper = DocNamespaceHelper()  # use my namespaces
        _proposedAlertStateNodes = [p.mkStateNode(msgTag('ProposedAlertState')) for p in _proposedAlertStates]

        return self._mkSetMethodSoapEnvelope('SetAlertState', operationHandle, _proposedAlertStateNodes)

    def _mkSetMetricStateEnvelope(self, operationHandle, proposedMetricStates):
        """create soap envelope, but do not send it. Used for unit testing
        :param proposedMetricState: a list of AbstractMetricStateContainer or derived classes """
        _proposedMetricStates = [p.mkCopy() for p in proposedMetricStates]
        nsmapper = DocNamespaceHelper()
        for p in _proposedMetricStates:
            p.nsmapper = nsmapper  # use my namespaces
        _proposedMetricStateNodes = [p.mkStateNode(msgTag('ProposedMetricState')) for p in _proposedMetricStates]

        return self._mkSetMethodSoapEnvelope('SetMetricState', operationHandle, _proposedMetricStateNodes)

    def _mkSetComponentStateEnvelope(self, operationHandle, proposedComponentStates):
        """Create soap envelope, but do not send it. Used for unit testing
        :param proposedComponentStates: a list of AbstractComponentStateContainers or derived classes """
        _proposedComponentStates = [p.mkCopy() for p in proposedComponentStates]
        nsmapper = DocNamespaceHelper()
        for p in _proposedComponentStates:
            p.nsmapper = nsmapper  # use my namespaces
        _proposedComponentStateNodes = [p.mkStateNode(msgTag('ProposedComponentState')) for p in
                                        _proposedComponentStates]

        return self._mkSetMethodSoapEnvelope('SetComponentState', operationHandle, _proposedComponentStateNodes)


class CTreeServiceClient(HostedServiceClient):

    def getDescriptorNode(self, handles, request_manipulator=None):
        """

        :param handles: a list of strings
        :return: a list of etree nodes
        """
        handle_nodes = []
        for h in handles:
            node = etree_.Element(msgTag('HandleRef'))
            node.text = h
            handle_nodes.append(node)
        resultSoapEnvelope = self._callGetMethod('GetDescriptor', params=handle_nodes,
                                                 request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode

    def getContainmentTreeNodes(self, handles, request_manipulator=None):
        """

        :param handles: a list of strings
        :return: a list of etree nodes
        """
        handle_nodes = []
        for h in handles:
            node = etree_.Element(msgTag('HandleRef'))
            node.text = h
            handle_nodes.append(node)
        resultSoapEnvelope = self._callGetMethod('GetContainmentTree', params=handle_nodes,
                                                 request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode


class StateEventClient(HostedServiceClient):
    subscribeable_actions = ('EpisodicMetricReport',
                             'EpisodicAlertReport',
                             'EpisodicComponentReport',
                             'EpisodicOperationalStateReport',
                             'PeriodicMetricReport',
                             'PeriodicAlertReport',
                             'PeriodicComponentReport',
                             'PeriodicOperationalStateReport'
                             )


class DescriptionEventClient(HostedServiceClient):
    subscribeable_actions = ('DescriptionModificationReport',)


class ContextServiceClient(HostedServiceClient):
    subscribeable_actions = ('EpisodicContextReport', 'PeriodicContextReport')

    def mkProposedContextObject(self, descriptorHandle, handle=None):
        """
        Helper method that create a state that can be used in setContextState operation
        :param descriptorHandle: the descriptor for which a state shall be created or updated
        :param handle: if None, a new object with default values is created (INSERT operation).
                       Otherwise a copy of an existing state with this handle is returned.
        :return: a context state instance
        """
        mdib = self._mdib_wref()
        if mdib is None:
            raise RuntimeError('no mdib information')
        contextDescriptorContainer = mdib.descriptions.handle.getOne(descriptorHandle)
        if handle is None:
            cls = self._sdc_definitions.sc.getContainerClass(contextDescriptorContainer.STATE_QNAME)
            obj = cls(nsmapper=DocNamespaceHelper(), descriptorContainer=contextDescriptorContainer)
            obj.Handle = descriptorHandle # this indicates that this is a new context state
        else:
            _obj = mdib.contextStates.handle.getOne(handle)
            obj = _obj.mkCopy()
        return obj

    def setContextState(self, operationHandle, proposedContextStates, request_manipulator=None):
        """
        """
        tmp = ', '.join(['{}(descriptorHandle={}, handle={})'.format(st.__class__.__name__,
                                                                     st.descriptorHandle,
                                                                     st.Handle)
                         for st in proposedContextStates])
        self._logger.info('setContextState {}', tmp)
        soapEnvelope = self._mkSetContextStateEnvelope(operationHandle, proposedContextStates)
        futureObject = self._callOperation(soapEnvelope, request_manipulator=request_manipulator)
        return futureObject

    def _mkSetContextStateEnvelope(self, operationHandle, proposedContextStates):
        """create soap envelope, but do not send it. Used for unit testing
        :param proposedContextStates: a list AbstractContextState or derived class """
        _proposedContextStates = [p.mkCopy() for p in proposedContextStates]
        for p in _proposedContextStates:
            # BICEPS: if handle == descriptorHandle, it means insert.
            if p.Handle is None:
                p.Handle = p.DescriptorHandle
            p.nsmapper = DocNamespaceHelper()  # use my namespaces
        _proposedContextStateNodes = [p.mkStateNode(msgTag('ProposedContextState')) for p in _proposedContextStates]

        return self._mkSetMethodSoapEnvelope('SetContextState', operationHandle, _proposedContextStateNodes)


    def getContextStatesNode(self, handles=None, request_manipulator=None):
        """
        :param handles: a list of handles
        """
        params = []
        if handles:
            for h in handles:
                params.append(etree_.Element(msgTag('HandleRef'), attrib={QN_TYPE: '{}:HandleRef'.format(Prefix.MSG.prefix)},
                                             nsmap=Prefix.partialMap(Prefix.MSG, Prefix.PM)))
                params[-1].text = h
        resultSoapEnvelope = self._callGetMethod('GetContextStates', params, request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode

    def getContextStateByIdentification(self, identifications, contextType=None, request_manipulator=None):
        """
        :param identifications: list of identifiers (type: InstanceIdentifier from pmtypes)
        :param contextType: Type to query
        :return:
        """
        params = []
        for oneId in identifications:
            params.append(oneId.asEtreeNode(qname=namespaces.msgTag('Identification'), nsmap=namespaces.nsmap))
        # todo: set attribute type based on contextType if set
        resultSoapEnvelope = self._callGetMethod('GetContextStatesByIdentification', params, request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode

class WaveformClient(HostedServiceClient):
    subscribeable_actions = ('Waveform',)




from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import domTag, siTag, nsmap
from .. import sdcdevice
from .. import pmtypes
from .. import safety
from ..nomenclature import NomenclatureCodes as nc

from . import providerbase


# coded values for SDC audio pause
MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE = pmtypes.CodedValue(nc.MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE)
MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE = pmtypes.CodedValue(nc.MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE)

class GenericSDCAudioPauseProvider(providerbase.ProviderRole):
    """Handling of global audio pause.
    It guarantees that there are operations with codes "MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE"
    and "MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE".
    """
    USE_SAFETYCONTEXT = False # using safetycontext can be enabled with this switch

    def __init__(self, log_prefix):
        super(GenericSDCAudioPauseProvider, self).__init__(log_prefix)
        self._setGlobalAudioPauseOperations = []
        self._cancelGlobalAudioPauseOperations = []

    def makeOperationInstance(self, operationDescriptorContainer):
        if operationDescriptorContainer.coding == MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE.coding:
            self._logger.info('instantiating "set audio pause" operation from existing descriptor handle={}'.format(operationDescriptorContainer.handle))
            set_ap_operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                        currentRequestHandler=self._setGlobalAudioPause)

            set_ap_operation.safetyReq = operationDescriptorContainer.SafetyReq
            self._setGlobalAudioPauseOperations.append(set_ap_operation)
            return set_ap_operation

        elif operationDescriptorContainer.coding == MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE.coding:
            self._logger.info('instantiating "cancel audio pause" operation from existing descriptor handle={}'.format(operationDescriptorContainer.handle))
            cancel_ap_operation = self._mkOperationFromOperationDescriptor(operationDescriptorContainer,
                                                                           currentRequestHandler=self._cancelGlobalAudioPause)

            self._cancelGlobalAudioPauseOperations.append(cancel_ap_operation)
            return cancel_ap_operation
        return None


    def makeMissingOperations(self):
        ops = []
        operationTargetContainer = self._mdib.descriptions.NODETYPE.getOne(
            domTag('MdsDescriptor'))  # the operation target is the mds itself
        if not self._setGlobalAudioPauseOperations:
            self._logger.info('adding "set audio pause" operation, no descriptor in mdib (looked for code = {})'.format(
                nc.MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE))
            set_ap_operation = self._mkOperation(sdcdevice.sco.ActivateOperation,
                                                 handle='AP__ON',
                                                 operationTargetHandle=operationTargetContainer.handle,
                                                 codedValue=MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE,
                                                 currentRequestHandler=self._setGlobalAudioPause)
            if self.USE_SAFETYCONTEXT:
                self._setAudioPauseSafetyRequirement(set_ap_operation)

            self._setGlobalAudioPauseOperations.append(set_ap_operation)
            ops.append(set_ap_operation)
        if not self._cancelGlobalAudioPauseOperations:
            self._logger.info(
                'adding "cancel audio pause" operation, no descriptor in mdib (looked for code = {})'.format(
                    nc.MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE))
            cancel_ap_operation = self._mkOperation(sdcdevice.sco.ActivateOperation,
                                                    handle='AP__CANCEL',
                                                    operationTargetHandle=operationTargetContainer.handle,
                                                    codedValue=MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE,
                                                    currentRequestHandler=self._cancelGlobalAudioPause)
            if self.USE_SAFETYCONTEXT:
                self._setAudioPauseSafetyRequirement(
                    cancel_ap_operation)  # risk wise not really needed, but S31 has also sefety context here.
            ops.append(cancel_ap_operation)
            self._setGlobalAudioPauseOperations.append(cancel_ap_operation)
        return ops

    def _checkSafetyContext(self, request):
        safetyInfoNode = request.headerNode.find(siTag('SafetyInfo'))
        if safetyInfoNode is None:
            raise Exception('No SafetyInfo in Header found')
        safetyContextNode = safetyInfoNode.find(siTag('SafetyContext'))
        if safetyContextNode is None:
            raise Exception('No SafetyContext in SafetyInfo Header found')
        locContextDescrContainer = self._mdib.descriptions.NODETYPE.getOne(domTag('LocationContextDescriptor'),
                                                                           allowNone=True)
        if locContextDescrContainer is None:
            raise Exception('Device has no LocationContext Descriptor instance')
        locationContextStateContainers = self._mdib.contextStates.NODETYPE.get(
            domTag('LocationContextState'))
        assoc = [l for l in locationContextStateContainers if
                 l.ContextAssociation == pmtypes.ContextAssociation.ASSOCIATED]
        if len(assoc) != 1:
            raise Exception('Expected one associated Location, got {}'.format(len(assoc)))
        extension = safetyContextNode.xpath('si:CtxtValue[@ReferencedSelector="LocationIIExt"]/text()',
                                            namespaces=nsmap)
        if assoc[0].Identification[0].Extension != extension[0]:
            raise Exception(
                'wrong SafetyContext Extension, expect "{}", got "{}"'.format(assoc[0].Identification.Extension,
                                                                              extension[0]))
        root = safetyContextNode.xpath('si:CtxtValue[@ReferencedSelector="LocationIIRoot"]/text()',
                                       namespaces=nsmap)
        if assoc[0].Identification[0].Root != root[0]:
            raise Exception(
                'wrong SafetyContext Root, expect "{}", got "{}"'.format(assoc[0].Identification.Root, root[0]))


    def _setAudioPauseSafetyRequirement(self, ap_operation):
        # create SafetyReq Definition
        dom_prefix = self._mdib.nsmapper.domPrefix()

        dualChannelDef = None

        locContextDescrContainer = self._mdib.descriptions.NODETYPE.getOne(domTag('LocationContextDescriptor'),
                                                                     allowNone=True)
        if locContextDescrContainer is not None:
            desc_h = locContextDescrContainer.handle
            safetyContextDef = pmtypes.T_SafetyContextDef([
                pmtypes.T_Selector('LocationIIExt',
                                   '/{pm}:MdState/{pm}:State[@DescriptorHandle="{dh}"][@ContextAssociation="Assoc"]/{pm}:Identification[1]/@Extension'.format(
                                       pm=dom_prefix, dh=desc_h)),
                pmtypes.T_Selector('LocationIIRoot',
                                   '/{pm}:MdState/{pm}:State[@DescriptorHandle="{dh}"][@ContextAssociation="Assoc"]/{pm}:Identification[1]/@Root'.format(
                                       pm=dom_prefix, dh=desc_h)),
            ])
        else:
            safetyContextDef = pmtypes.T_SafetyContextDef([])
        ap_operation.safetyReq = pmtypes.T_SafetyReq(dualChannelDef=dualChannelDef, safetyContextDef=safetyContextDef)


    def _setGlobalAudioPause(self, operationInstance, request):  # pylint: disable=unused-argument
        """ This is the code that executes the operation itself:
        SF1132: If global audio pause is initiated, all SystemSignalActivation/State for all alarm systems of the
        product with SystemSignalActivation/Manifestation evaluating to 'Aud' shall be set to 'Psd'.

        SF958: If signal pause is initiated for an alert signal that is not an ACKNOWLEDGE CAPABLE ALERT SIGNAL,
        then the Alert Provider shall set the AlertSignalState/ActivationState to 'Psd' and the AlertSignalState/Presence to 'Off'.

        SF959: If signal pause is initiated for an ACKNOWLEDGEABLE ALERT SIGNAL, the the Alert Provider shall set the
        AlertSignalState/ActivationState to 'Psd' and AlertSignalState/Presence to 'Ack' for that ALERT SIGNAL.
         """
        alertSystemDescriptors = self._mdib.descriptions.NODETYPE.get(domTag('AlertSystemDescriptor'))
        if alertSystemDescriptors is None:
            self._logger.error('SDC_SetAudioPauseOperation called, but no AlertSystemDescriptor in mdib found')
            return
        if self.USE_SAFETYCONTEXT:
            self._checkSafetyContext(request)

        with self._mdib.mdibUpdateTransaction() as tr:
            for alertSystemDescriptor in alertSystemDescriptors:
                alertSystemState = tr.getAlertState(alertSystemDescriptor.handle)
                if alertSystemState.ActivationState != pmtypes.AlertActivation.ON:
                    self._logger.info('SDC_SetAudioPauseOperation: nothing to do')
                    tr.ungetState(alertSystemState)
                else:
                    audible_signals = [ ssa for ssa in alertSystemState.SystemSignalActivation if ssa.Manifestation == pmtypes.AlertSignalManifestation.AUD]
                    active_audible_signals = [ ssa for ssa in audible_signals if ssa.State != pmtypes.AlertActivation.PAUSED]
                    if not active_audible_signals:
                        # Alert System has no audible SystemSignalActivations, no action required
                        tr.ungetState(alertSystemState)
                    else:
                        for ssa in active_audible_signals:
                            ssa.State = pmtypes.AlertActivation.PAUSED # SF1132
                        self._logger.info('SDC_SetAudioPauseOperation: set alertsystem "{}" to paused'.format(
                            alertSystemDescriptor.handle))
                        # handle all audible alert signals of this alert system
                        allAlertSignalDescriptors = self._mdib.descriptions.NODETYPE.get(domTag('AlertSignalDescriptor'), [])
                        childAlertSignalDescriptors = [ d for d in allAlertSignalDescriptors if d.parentHandle == alertSystemDescriptor.handle]
                        audibleChildAlertSignalDescriptors = [ d for d in childAlertSignalDescriptors if d.Manifestation == pmtypes.AlertSignalManifestation.AUD]
                        for sd in audibleChildAlertSignalDescriptors:
                            alertSignalState = tr.getAlertState(sd.handle)
                            if sd.AcknowledgementSupported: #SF959
                                if alertSignalState.ActivationState != pmtypes.AlertActivation.PAUSED \
                                    or alertSignalState.Presence != pmtypes.AlertSignalPresence.ACK:
                                    alertSignalState.ActivationState = pmtypes.AlertActivation.PAUSED
                                    alertSignalState.Presence = pmtypes.AlertSignalPresence.ACK
                                else:
                                    tr.ungetState(alertSignalState)
                            else: #SF958
                                if alertSignalState.ActivationState != pmtypes.AlertActivation.PAUSED \
                                    or alertSignalState.Presence != pmtypes.AlertSignalPresence.OFF:
                                    alertSignalState.ActivationState = pmtypes.AlertActivation.PAUSED
                                    alertSignalState.Presence = pmtypes.AlertSignalPresence.OFF
                                else:
                                    tr.ungetState(alertSignalState)


    def _cancelGlobalAudioPause(self, operationInstance, request): #pylint: disable=unused-argument
        """ This is the code that executes the operation itself:
        If global audio pause is initiated, all SystemSignalActivation/State for all alarm systems of the product with
        SystemSignalActivation/Manifestation evaluating to 'Aud' shall be set to 'Psd'.
         """
        alertSystemDescriptors = self._mdib.descriptions.NODETYPE.get(domTag('AlertSystemDescriptor'))
        if self.USE_SAFETYCONTEXT:
            self._checkSafetyContext(request)
        with self._mdib.mdibUpdateTransaction() as tr:
            for alertSystemDescriptor in alertSystemDescriptors:
                alertSystemState = tr.getAlertState(alertSystemDescriptor.handle)
                if alertSystemState.ActivationState != pmtypes.AlertActivation.ON:
                    self._logger.info('SDC_CancelAudioPauseOperation: nothing to do')
                    tr.ungetState(alertSystemState)
                else:
                    audible_signals = [ ssa for ssa in alertSystemState.SystemSignalActivation if ssa.Manifestation == pmtypes.AlertSignalManifestation.AUD]
                    paused_audible_signals = [ ssa for ssa in audible_signals if ssa.State == pmtypes.AlertActivation.PAUSED]
                    if not paused_audible_signals:
                        tr.ungetState(alertSystemState)
                    else:
                        for ssa in paused_audible_signals:
                            ssa.State = pmtypes.AlertActivation.ON
                        self._logger.info('SDC_SetAudioPauseOperation: set alertsystem "{}" to ON'.format(
                            alertSystemDescriptor.handle))
                        # handle all audible alert signals of this alert system
                        allAlertSignalDescriptors = self._mdib.descriptions.NODETYPE.get(domTag('AlertSignalDescriptor'), [])
                        childAlertSignalDescriptors = [ d for d in allAlertSignalDescriptors if d.parentHandle == alertSystemDescriptor.handle]
                        audibleChildAlertSignalDescriptors = [ d for d in childAlertSignalDescriptors if d.Manifestation == pmtypes.AlertSignalManifestation.AUD]
                        for sd in audibleChildAlertSignalDescriptors:
                            alertSignalState = tr.getAlertState(sd.handle)
                            alertConditionState = self._mdib.states.descriptorHandle.getOne(sd.ConditionSignaled)
                            if alertConditionState.Presence:
                                # set signal back to 'ON'
                                if alertSignalState.ActivationState == pmtypes.AlertActivation.PAUSED:
                                    alertSignalState.ActivationState = pmtypes.AlertActivation.ON
                                    alertSignalState.Presence = pmtypes.AlertSignalPresence.ON
                                else:
                                    tr.ungetState(alertSignalState)



from collections import defaultdict, namedtuple
from threading import RLock

"""
This module implements an in-memory table with indices for faster access to objects.
Example: You have a class 
class Person:
    def __init__(self, first_name, last_name, age):
        self.first_name = first_name
        self.last_name = last_name
        self._age = age

    def getAge(self):
        return int(self._age

you can setup an in-memory table like this:
person_lookup = MultiKeyLookup()
person_lookup.addIndex('by_firstname', multikey.IndexDefinition(lambda obj: obj.first_name))
person_lookup.addIndex('by_firstname', multikey.IndexDefinition(lambda obj: obj.last_name))
person_lookup.addIndex('by_age', multikey.IndexDefinition(lambda obj: obj.getAge()))

person_lookup.addObject(Person('Peter', 'Miller', 42)
person_lookup.addObject(Person('John', 'Myers', 50)
person_lookup.addObject(Person('Agnes', 'Miller', 42)

accessing by index:
all_millers = person_lookup.by_lastname.get('Miller')
all_42_agers = person_lookup.by_age.get(42)
"""


class IndexDefinition(dict):
    """ An index allows to group objects by values.
    This is a dictionary that has lists ob objects as value.
    Each list contains objects that have the same key member"""

    def __init__(self, getKeyFunc, indexNoneValues=True):
        """
        :param getKeyFunc: a callable that returns a key value from a given object
        :param indexNoneValues: if True, a None key is handled like every other value.
                                if False,a None key is not added to index.
        """
        super(IndexDefinition, self).__init__()
        self._getKeyFunc = getKeyFunc
        self._indexNoneValues = indexNoneValues

    def getOne(self, key, allowNone=False):
        try:
            result = self[key]
            if len(result) > 1:
                raise RuntimeError('getOne: key "{}" has {} objects'.format(key, len(result)))
            return result[0]
        except KeyError:
            if allowNone:
                return
            raise RuntimeError('key "{}" not found'.format(key))

    def _mkKeys(self, obj):
        key = self._getKeyFunc(obj)
        if not self._indexNoneValues and key is None:
            return
        try:
            self[key].append(obj)
        except KeyError:
            self[key] = [obj]
        return [key]

    def _rmKey(self, key, obj):
        try:
            objList = self[key]
            objList.remove(obj)
            if len(objList) == 0:
                del self[key]
        except (KeyError, ValueError):
            pass


class UIndexDefinition(IndexDefinition):
    """ A unique Index, there can only be one object with that key"""

    def _mkKeys(self, obj):
        keys = self._getKeyFunc(obj)
        if not self._indexNoneValues and keys is None:
            return
        if isinstance(keys, list):
            raise ValueError('list of keys not allowed in UIndex: obj= {}, keys={}'.format(obj, keys))
            pass
        else:
            keys = [keys]
        for k in keys:
            if k in self:
                raise KeyError('key "{}" in already in this UIndex'.format(k))
            self[k] = [obj]
        return keys


class IndexDefinition1n(IndexDefinition):
    """ For member values that are a list of keys (1:n relationship)"""

    def _mkKeys(self, obj):
        keys = self._getKeyFunc(obj)
        if not self._indexNoneValues and keys is None:
            return
        for k in keys:
            try:
                self[k].append(obj)
            except KeyError:
                self[k] = [obj]
        return keys


class ObjectSelector(object):
    def __init__(self, selectedObjects):
        self.objects = selectedObjects

    def find(self, **kwargs):
        """ OR combination of args. Values are compared for equality (==), not identity (is)."""
        result = []
        for o in self.objects:
            for name, value in kwargs.items():
                try:
                    val = getattr(o, name)
                    if callable(val):
                        val = val()
                except AttributeError:
                    pass
                else:
                    if val == value:
                        result.append(o)
                        break
        return ObjectSelector(result)


_ObjRef = namedtuple('_ObjRef', 'index_dict key')  # used internally in MultiKeyLookup to keep track of all indexes.


# when we remove an object we need it to delete all indices referencing it

class MultiKeyLookup(object):

    def __init__(self):
        self._objects = set()  # contains the objects
        self._objectIDs = defaultdict(
            list)  # key = id(object), value = list of ((_idxDefs, key) tuples that reference the object
        self._idxDefs = {}  # holds UIndexDefinition Objects
        self._lock = RLock()

    @property
    def objects(self):
        return self._objects

    @property
    def lock(self):
        return self._lock

    def getIndexDict(self, indexName):
        return self._idxDefs[indexName]  # .indices

    def __getattr__(self, name):
        return self._idxDefs[name]  # .indices

    def addIndex(self, indexName, indexDefinition):
        self._idxDefs[indexName] = indexDefinition
        # add existing objects to new lookup
        for obj in self._objects:
            keys = indexDefinition._mkKeys(obj)
            for k in keys:
                self._objectIDs[id(obj)].append(_ObjRef(indexDefinition, k))

    def addObject(self, obj):
        if obj in self._objects:
            return
        with self._lock:
            self._objects.add(obj)
            self._mkIndices(obj)

    def addObjectNoLock(self, obj):
        if obj in self._objects:
            return
        self._objects.add(obj)
        self._mkIndices(obj)

    def addObjects(self, objs):
        with self._lock:
            self.addObjectsNoLock(objs)

    def addObjectsNoLock(self, objs):
        for obj in objs:
            if obj in self._objects:
                continue
            self._objects.add(obj)
            self._mkIndices(obj)

    def _mkIndices(self, obj):
        all_keys = []  # for this object
        for indexDefinition in self._idxDefs.values():
            try:
                keys = indexDefinition._mkKeys(obj)
                for k in keys:
                    all_keys.append(_ObjRef(indexDefinition, k))
            except (TypeError, AttributeError):
                pass
        self._objectIDs[id(obj)].extend(all_keys)

    def _rmIndices(self, obj):
        obj_refs = self._objectIDs.get(id(obj), [])
        for obj_ref in obj_refs:
            obj_ref.index_dict._rmKey(obj_ref.key, obj)
        del self._objectIDs[id(obj)]

    def removeObject(self, obj):
        obj_refs = self._objectIDs.get(id(obj))
        if obj_refs is None:
            return
        with self._lock:
            self._rmIndices(obj)
            self._objects.remove(obj)

    def removeObjectNoLock(self, obj):
        obj_refs = self._objectIDs.get(id(obj))
        if obj_refs is None:
            return
        self._rmIndices(obj)
        self._objects.remove(obj)

    def removeObjects(self, objs):
        with self._lock:
            self.removeObjectsNoLock(objs)

    def removeObjectsNoLock(self, objs):
        for obj in objs:
            obj_refs = self._objectIDs.get(id(obj))
            if obj_refs is None:
                continue
            self._rmIndices(obj)
            self._objects.remove(obj)

    def updateObject(self, obj):
        if obj not in self._objects:
            raise RuntimeError('object {} not known'.format(obj))
        with self._lock:
            self._rmIndices(obj)
            self._mkIndices(obj)

    def updateObjectNoLock(self, obj):
        if obj not in self._objects:
            raise RuntimeError('object {} not known'.format(obj))
        self._rmIndices(obj)
        self._mkIndices(obj)

    def updateObjects(self, objs):
        with self._lock:
            self.updateObjectsNoLock(objs)

    def updateObjectsNoLock(self, objs):
        for obj in objs:
            if obj not in self._objects:
                raise RuntimeError('object {} not known'.format(obj))
            with self._lock:
                self._rmIndices(obj)
                self._mkIndices(obj)

    def clear(self):
        with self._lock:
            for indexDefinition in self._idxDefs.values():
                indexDefinition.clear()
            self._objectIDs.clear()
            self._objects.clear()

    def find(self, **kwargs):
        sel = ObjectSelector(self._objects)
        with self._lock:
            return sel.find(**kwargs)

    def findNoLock(self, **kwargs):
        sel = ObjectSelector(self._objects)
        return sel.find(**kwargs)




""" A helper for xml name space handling"""
from lxml import etree as etree_
from _functools import partial
from collections import namedtuple

_Prefix_Namespace_Tuple = namedtuple('_Prefix_Namespace_Tuple', 'prefix namespace')

from enum import Enum

# these are internally used namespaces, they are not all identical the ones that are used in sdc.
# it is in the responsibility of the sdc definitions class(es) to convert between internal and external namespaces.
# Originally this abstraction was needed, because during development of the sdc standard the namespaces changed.
# Although the standard is meanwhile final, this abstraction might again be needed if in the future a new revision
# of the standard appears.
class Prefix_Namespace(_Prefix_Namespace_Tuple, Enum):
    MSG = _Prefix_Namespace_Tuple('msg', "__BICEPS_MessageModel__")
    PM = _Prefix_Namespace_Tuple('dom', "__BICEPS_ParticipantModel__")
    MDPWS = _Prefix_Namespace_Tuple('mdpws', "__MDPWS__")
    EXT = _Prefix_Namespace_Tuple('ext', "__ExtensionPoint__")
    SDC = _Prefix_Namespace_Tuple('sdc', 'http://standards.ieee.org/downloads/11073/11073-20701-2018')
    WSE = _Prefix_Namespace_Tuple('wse', 'http://schemas.xmlsoap.org/ws/2004/08/eventing')
    XSD = _Prefix_Namespace_Tuple('xsd', 'http://www.w3.org/2001/XMLSchema')
    XSI = _Prefix_Namespace_Tuple('xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    WSA = _Prefix_Namespace_Tuple('wsa', 'http://www.w3.org/2005/08/addressing')
    WSX = _Prefix_Namespace_Tuple('wsx', 'http://schemas.xmlsoap.org/ws/2004/09/mex') # Meta Data Exchange
    DPWS = _Prefix_Namespace_Tuple('dpws', 'http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01')
    WSD = _Prefix_Namespace_Tuple('wsd', 'http://docs.oasis-open.org/ws-dd/ns/discovery/2009/01')
    S12 = _Prefix_Namespace_Tuple('s12', 'http://www.w3.org/2003/05/soap-envelope')
    XML = _Prefix_Namespace_Tuple('xml', 'http://www.w3.org/XML/1998/namespace')
    WXF = _Prefix_Namespace_Tuple('wxf', 'http://schemas.xmlsoap.org/ws/2004/09/transfer') # ws-transfer
    WSDL = _Prefix_Namespace_Tuple('wsdl', 'http://schemas.xmlsoap.org/wsdl/')
    @staticmethod
    def partialMap(*prefix):
        """
        :param prefix: Prefix_Namespace_Tuples
        :return: a dictionary with prefix as key, namespace as value
        """
        return dict((v.prefix, v.namespace) for v in prefix)


# these are all namespaces used in sdc:
nsmap = dict((item.prefix, item.namespace) for item in Prefix_Namespace)

def _tagName(prefix_namespace_tuple, tagname):
    return etree_.QName(prefix_namespace_tuple.namespace, tagname)

msgTag = partial(_tagName, Prefix_Namespace.MSG) # a helper to make qualified names
domTag = partial(_tagName, Prefix_Namespace.PM) # a helper to make qualified names
extTag = partial(_tagName, Prefix_Namespace.EXT) # a helper to make qualified names
wseTag = partial(_tagName, Prefix_Namespace.WSE) # a helper to make qualified names
xsiTag = partial(_tagName, Prefix_Namespace.XSI) # a helper to make qualified names
wsaTag = partial(_tagName, Prefix_Namespace.WSA) # a helper to make qualified names
wsxTag = partial(_tagName, Prefix_Namespace.WSX) # a helper to make qualified names
dpwsTag = partial(_tagName, Prefix_Namespace.DPWS) # a helper to make qualified names
mdpwsTag = partial(_tagName, Prefix_Namespace.MDPWS) # a helper to make qualified names
siTag  = partial(_tagName, Prefix_Namespace.MDPWS) # a helper to make qualified names
wsdTag = partial(_tagName, Prefix_Namespace.WSD) # a helper to make qualified names
s12Tag = partial(_tagName, Prefix_Namespace.S12) # a helper to make qualified names
xmlTag = partial(_tagName, Prefix_Namespace.XML) # a helper to make qualified names


# some constants from ws-addressing
WSA_ANONYMOUS = Prefix_Namespace.WSA.namespace + '/anonymous'
WSA_NONE = Prefix_Namespace.WSA.namespace + '/none'
WSA_IS_REFERENCE_PARAMETER = '{' + Prefix_Namespace.WSA.namespace + '}IsReferenceParameter'


def docNameFromQName(qName, ns_map):
    """ returns the docprefix:name string, or only name (if default namespace is used) """
    prefixmap = dict ((v,k) for k, v in ns_map.items())
    prefix = prefixmap[qName.namespace]
    if prefix is None:
        return qName.localname
    else:
        return '{}:{}'.format(prefix, qName.localname)


class DocNamespaceHelper(object):
    def __init__(self):
        self._prefixmap = dict ((x.namespace, x.prefix) for x in Prefix_Namespace)

    def useDocPrefixes(self, document_nsmap):
        for prefix, _ns in document_nsmap.items():
            self._prefixmap[_ns] = prefix

    def _docPrefix(self, prefix_namespace_tuple):
        """ returns the document prefix for nsmap prefix"""
        return self._prefixmap[prefix_namespace_tuple.namespace]

    def msgPrefix(self):
        return self._prefixmap[Prefix_Namespace.MSG.namespace]

    def domPrefix(self):
        return self._prefixmap[Prefix_Namespace.PM.namespace]

    def docName(self, myPrefix_or_namespace, name):
        """ returns the docprefix:name string. """
        prefix = self._docPrefix(myPrefix_or_namespace)
        if prefix is None:
            return name
        else:
            return '{}:{}'.format(prefix, name)

    def docNameFromQName(self, qName):
        """ returns the docprefix:name string, or only name (if default namespace is used) """
        prefix = self._prefixmap[qName.namespace]
        if prefix is None:
            return qName.localname
        else:
            return '{}:{}'.format(prefix, qName.localname)

    @property
    def docNssmap(self):
        return dict ((v,k) for k, v in self._prefixmap.items())

    def partialMap(self, *prefix):
        """
        :param prefix: Prefix class members
        :return: a dictionary with prefix as key, namespace as value
        """
        mynamespaces = []
        for p in prefix:
            if isinstance(p, _Prefix_Namespace_Tuple):
                mynamespaces.append(p.namespace)
            else:
                mynamespaces.append(nsmap[p])
        return dict ((v,k) for k, v in self._prefixmap.items() if k in mynamespaces)


QN_TYPE = xsiTag('type')   # frequently used QName, central definition


def txt2QName(text, doc_nsmap):
    elements = text.split(':')
    prefix = None if len(elements) == 1 else elements[0]
    name = elements[-1]
    try:
        return etree_.QName(doc_nsmap[prefix], name)
    except KeyError:
        raise KeyError('Cannot make QName for {}, prefix is not in nsmap: {}'.format(text, doc_nsmap.keys()))



from sdc11073 import isoduration
from decimal import Decimal

class NullConverter(object):
    @staticmethod
    def toPy(xmlValue):
        return xmlValue
    @staticmethod
    def toXML(pyValue):
        return pyValue



class TimestampConverter(object):
    """ XML representation: integer, representing timestamp in milliseconds
     Python representation: float in seconds
    """
    @staticmethod
    def toPy(xmlValue):
        return float(xmlValue) / 1000.0
    @staticmethod
    def toXML(pyValue):
        ms_value = int(pyValue*1000)
        return str(ms_value)


class DecimalConverter(object):
    USE_DECIMAL_TYPE = True

    @classmethod
    def toPy(cls, xmlValue):
        if cls.USE_DECIMAL_TYPE:
            return Decimal(xmlValue)
        else:
            if '.' in xmlValue:
                return float(xmlValue)
            else:
                return int(xmlValue)

    @staticmethod
    def toXML(pyValue):
        if isinstance(pyValue, float):
            # round value to handle float inaccuracies
            if abs(pyValue) >= 100:
                xmlValue = '{:.1f}'.format(round(pyValue, 1))
            elif abs(pyValue) >= 10:
                xmlValue = '{:.2f}'.format(round(pyValue, 2))
            else:
                xmlValue = '{:.3f}'.format(round(pyValue, 3))
        elif isinstance(pyValue, Decimal):
            # assume Decimal is exact, no rounding errors
            # Decimal has no method to force string representation without exponential notion.
            # => convert to float and use :f string formatting (6 digits after decimal point, which should be good enough)
            xmlValue = f'{pyValue:f}'
        else:
            xmlValue = str(pyValue)
        # remove trailing zeros after decimal point
        while '.' in xmlValue and xmlValue[-1] in ('0', '.'):
            xmlValue = xmlValue[:-1]
        return xmlValue


class IntegerConverter(object):
    @staticmethod
    def toPy(xmlValue):
        return int(xmlValue)
    @staticmethod
    def toXML(pyValue):
        return str(pyValue)



class BooleanConverter(object):
    @staticmethod
    def toPy(xmlValue):
        return xmlValue == 'true'
    @staticmethod
    def toXML(pyValue):
        if pyValue:
            return 'true'
        else:
            return 'false'


class DurationConverter(object):
    @staticmethod
    def toPy(xmlValue):
        if xmlValue is None:
            return None
        return isoduration.parse_duration(xmlValue)
    @staticmethod
    def toXML(pyValue):
        return isoduration.durationString(pyValue)



import os
from .namespaces import dpwsTag
from .namespaces import Prefix_Namespace as Prefix
schemaFolder = os.path.join(os.path.dirname(__file__), 'xsd')


class ProtocolsRegistry(type):
    """
    base class that has the only purpose to register classes that use this as meta class
    """
    protocols = []

    def __new__(cls, name, *arg, **kwarg):
        new_cls = super().__new__(cls, name, *arg, **kwarg)
        if name != 'BaseDefinitions': # ignore the base class itself
            cls.protocols.append(new_cls)
        return new_cls


# definitions that group all relevant dependencies for BICEPS versions
class BaseDefinitions(metaclass=ProtocolsRegistry):
    """ Central definitions for SDC
    It defines namespaces and handlers for the protocol.
    Derive from this class in order to define different protocol handling."""
    DpwsDeviceType = dpwsTag('Device')
    MetaDataExchangeSchemaFile = os.path.join(schemaFolder, 'MetadataExchange.xsd')
    EventingSchemaFile = os.path.join(schemaFolder, 'eventing.xsd')
    SoapEnvelopeSchemaFile = os.path.join(schemaFolder, 'soap-envelope.xsd')
    WsAddrSchemaFile = os.path.join(schemaFolder, 'ws-addr.xsd')
    EventingSchemaFile = os.path.join(schemaFolder, 'eventing.xsd')
    AddressingSchemaFile = os.path.join(schemaFolder, 'addressing.xsd')
    XMLSchemaFile = os.path.join(schemaFolder, 'xml.xsd')
    DPWSSchemaFile = os.path.join(schemaFolder, 'wsdd-dpws-1.1-schema-os.xsd')
    WSDiscoverySchemaFile = os.path.join(schemaFolder, 'wsdd-discovery-1.1-schema-os.xsd')
    WSDLSchemaFile = os.path.join(schemaFolder, 'wsdl.xsd')

    SchemaFilePaths = None

    # set the following namespaces in derived classes:
    MedicalDeviceTypeNamespace = None
    BICEPSNamespace = None
    MessageModelNamespace = None
    ParticipantModelNamespace = None
    ExtensionPointNamespace = None
    MedicalDeviceType = None
    ActionsNamespace = None

    @classmethod
    def ns_matches(cls, ns):
        """ This method checks if this definition set is the correct one for a given namespace"""
        return ns in (cls.MedicalDeviceTypeNamespace, cls.BICEPSNamespace, cls.MessageModelNamespace, cls.ParticipantModelNamespace, cls.ExtensionPointNamespace, cls.MedicalDeviceType)

    @classmethod
    def normalizeXMLText(cls, xml_text):
        """ replace BICEPS namespaces with internal namespaces"""
        for ns, internal_ns in ((cls.MessageModelNamespace, Prefix.MSG.namespace), #'__BICEPS_MessageModel__'),
                                (cls.ParticipantModelNamespace, Prefix.PM.namespace), #'__BICEPS_ParticipantModel__'),
                                (cls.ExtensionPointNamespace, Prefix.EXT.namespace), #'__ExtensionPoint__'),
                                (cls.MDPWSNameSpace, Prefix.MDPWS.namespace)): #'__MDPWS__')):
            xml_text = xml_text.replace('"{}"'.format(ns).encode('utf-8'), '"{}"'.format(internal_ns).encode('utf-8'))
        return xml_text

    @classmethod
    def denormalizeXMLText(cls, xml_text):
        """ replace internal namespaces with BICEPS namespaces"""
        for ns, internal_ns in ((cls.MessageModelNamespace.encode('utf-8'), b'__BICEPS_MessageModel__'),
                                (cls.ParticipantModelNamespace.encode('utf-8'), b'__BICEPS_ParticipantModel__'),
                                (cls.ExtensionPointNamespace.encode('utf-8'), b'__ExtensionPoint__'),
                                (cls.MDPWSNameSpace.encode('utf-8'), b'__MDPWS__')):
            xml_text = xml_text.replace(internal_ns, ns)
        return xml_text

    @classmethod
    def get_schema_file_path(cls, url):
        return cls.SchemaFilePaths.schema_location_lookup.get(url)



import weakref
from threading import Lock
from collections import namedtuple
from ..pmtypes import InvocationState
from .. import loghelper
from ..namespaces import nsmap
from concurrent.futures import Future


OperationResult = namedtuple('OperationResult', 'state error errorMsg soapEnvelope')


class OperationsManager(object):
    nonFinalOperationStates = (InvocationState.WAIT, InvocationState.START)
    def __init__(self, log_prefix):
        self.log_prefix = log_prefix
        self._logger = loghelper.getLoggerAdapter('sdc.client.op_mgr', log_prefix)
        self._transactions = {}
        self._transactionsLock = Lock()


    def callOperation(self, hostedServiceClient, soapEnvelope, request_manipulator=None):
        """ an operation call does not return the result of the operation directly. Instead you get an transaction id,
        and will receive the status of this transaction as notification ("OperationInvokedReport").
        This method returns a "future" object. The future object has a result as soon as a final transaction state is received.
        :param soapenvelope: the envelope to be sent
        @return: a concurrent.futures.Future object
        """
        ret = Future()
        with self._transactionsLock:
            resultSoapEnvelope = hostedServiceClient.postSoapEnvelope(soapEnvelope,
                                                                      msg='call Operation',
                                                                      request_manipulator=request_manipulator)
            transactionId = resultSoapEnvelope.msgNode.xpath('msg:InvocationInfo/msg:TransactionId/text()', namespaces=nsmap)[0]
            invocationState = resultSoapEnvelope.msgNode.xpath('msg:InvocationInfo/msg:InvocationState/text()', namespaces=nsmap)[0]

            if invocationState in self.nonFinalOperationStates:
                self._transactions[int(transactionId)] = weakref.ref(ret)
                self._logger.info('callOperation: transactionId {} registered, state={}', transactionId, invocationState)
            else:
                errors = resultSoapEnvelope.msgNode.xpath('msg:InvocationInfo/msg:InvocationError/text()', namespaces=nsmap)
                error = '' if len(errors) == 0 else str(errors[0])
                errorMsgs = resultSoapEnvelope.msgNode.xpath('msg:InvocationInfo/msg:InvocationErrorMessage/text()', namespaces=nsmap)
                errorMsg = '' if len(errorMsgs) == 0 else str(errorMsgs[0])

                result = OperationResult(invocationState, error, errorMsg, resultSoapEnvelope)
                self._logger.debug('Result of Operation: {}',  result)
                ret.set_result(result)
        return ret


    def onOperationInvokedReport(self, soapEnvelope):
        self._logger.debug('onOperationInvokedReport: response= {}', lambda: soapEnvelope.as_xml(pretty=True))
        transactionId = soapEnvelope.msgNode.xpath('msg:ReportPart/msg:InvocationInfo/msg:TransactionId/text()', namespaces=nsmap)[0]
        operationState = soapEnvelope.msgNode.xpath('msg:ReportPart/msg:InvocationInfo/msg:InvocationState/text()', namespaces=nsmap)[0]
        self._logger.debug('{}onOperationInvokedReport: got transactionId {} state {}', self.log_prefix, transactionId, operationState)
        if operationState in self.nonFinalOperationStates:
            self._logger.debug('nonFinal state detected, ignoring message...')
            return
        with self._transactionsLock:
            futureRef = self._transactions.pop(int(transactionId), None)
        if futureRef is None:
            # this was not my transaction
            self._logger.debug('transactionId {} is not registered!', transactionId)
            return
        futureObj = futureRef()
        if futureObj is None:
            # client gave up.
            self._logger.debug('transactionId {} given up', transactionId)
            return
        else:
            errors = soapEnvelope.msgNode.xpath('msg:ReportPart/msg:InvocationInfo/msg:InvocationError/text()', namespaces=nsmap)
            errorMsgs = soapEnvelope.msgNode.xpath('msg:ReportPart/msg:InvocationInfo/msg:InvocationErrorMessage/text()', namespaces=nsmap)
            result = OperationResult(operationState, ''.join(errors), ''.join(errorMsgs), soapEnvelope)
            if operationState == InvocationState.FAILED:
                self._logger.warn('transaction Id {} finished with error: error={}, error-message={}',
                                  transactionId, result.error, result.errorMsg)
            else:
                self._logger.info('transaction Id {} ok', transactionId)
            futureObj.set_result(result)




import sdc11073.certloader
from .sdc_handlers import SdcHandler_Full
from .. import pmtypes


class SdcDevice(object):
    defaultInstanceIdentifiers = (pmtypes.InstanceIdentifier(root='rootWithNoMeaning', extensionString='System'),)

    def __init__(self, ws_discovery, my_uuid, model, device, deviceMdibContainer, validate=True, roleProvider=None,
                 sslContext=None,
                 logLevel=None, max_subscription_duration=7200, log_prefix='', handler_cls=None,
                 chunked_messages=False,
                 ssl_context_container: sdc11073.certloader.SSLContextContainer = None):  # pylint:disable=too-many-arguments
        # ssl protocol handling itself is delegated to a handler.
        # Specific protocol versions or behaviours are implemented there.
        if handler_cls is None:
            handler_cls = SdcHandler_Full

        if sslContext and ssl_context_container:
            raise ValueError('sslContext and ssl_context_container must not both be given')

        if ssl_context_container:
            self._handler = handler_cls(my_uuid, ws_discovery, model, device, deviceMdibContainer, validate,
                                        roleProvider, sslContext, logLevel,
                                        max_subscription_duration, ssl_context_container=ssl_context_container,
                                        log_prefix=log_prefix, chunked_messages=chunked_messages)
        else:
            self._handler = handler_cls(my_uuid, ws_discovery, model, device, deviceMdibContainer, validate,
                                        roleProvider, sslContext, logLevel,
                                        max_subscription_duration,
                                        log_prefix=log_prefix, chunked_messages=chunked_messages)

        self._wsdiscovery = ws_discovery
        self._logger = self._handler._logger
        self._mdib = deviceMdibContainer
        self._location = None

    def setLocation(self, location, validators=defaultInstanceIdentifiers, publishNow=True):
        """
        :param location: a pysdc.location.SdcLocation instance
        :param validators: a list of pmtypes.InstanceIdentifier objects or None; in that case the defaultInstanceIdentifiers member is used
        :param publishNow: if True, the device is published via its wsdiscovery reference.
        """
        if location == self._location:
            return
        self._location = location
        self._mdib.setLocation(location, validators)
        if publishNow:
            self.publish()

    def publish(self):
        """
        publish device on the network (sends HELLO message)
        :return:
        """
        scopes = self._handler.mkScopes()
        xAddrs = self.getXAddrs()
        self._wsdiscovery.publishService(self.epr, self._mdib.sdc_definitions.MedicalDeviceTypesFilter, scopes, xAddrs)

    @property
    def shallValidate(self):
        return self._handler._validate

    @property
    def mdib(self):
        return self._mdib

    @property
    def subscriptionsManager(self):
        return self._handler._subscriptionsManager

    @property
    def scoOperationsRegistry(self):
        return self._handler._scoOperationsRegistry

    @property
    def epr(self):
        # End Point Reference, e.g 'urn:uuid:8c26f673-fdbf-4380-b5ad-9e2454a65b6b'
        return self._handler._my_uuid.urn

    @property
    def path_prefix(self):
        # http path prefix of service e.g '8c26f673-fdbf-4380-b5ad-9e2454a65b6b'
        return self._handler.path_prefix

    def registerOperation(self, operation):
        return self._handler.registerOperation(operation)

    def unRegisterOperationByHandle(self, operationHandle):
        return self._handler.unRegisterOperationByHandle(operationHandle)

    def getOperationByHandle(self, operationHandle):
        return self._handler.getOperationByHandle(operationHandle)

    def enqueueOperation(self, operation, request):
        return self._handler.enqueueOperation(operation, request)

    def dispatchGetRequest(self, parseResult, headers):
        """ device itself can also handle GET requests. This is the handler"""
        return self._handler.dispatchGetRequest(parseResult, headers)

    def startAll(self,
                 startRealtimeSampleLoop=True,
                 periodic_reports_interval=None,
                 shared_http_server=None,
                 http_server_timeout=15.0):
        """

        :param startRealtimeSampleLoop: flag
        :param periodic_reports_interval: if provided, a value in seconds
        :param shared_http_server: id provided, use this http server. Otherwise device creates its own.
        :param http_server_timeout: the timeout in seconds for waiting on the start event of
                                    the http server thread of the underlying handler
        :return:
        """
        return self._handler.startAll(startRealtimeSampleLoop, periodic_reports_interval, shared_http_server,
                                      http_server_timeout=http_server_timeout)

    def stopAll(self, closeAllConnections=True, sendSubscriptionEnd=True):
        return self._handler.stopAll(closeAllConnections, sendSubscriptionEnd)

    def getXAddrs(self):
        return self._handler.getXAddrs()

    def sendMetricStateUpdates(self, mdib_version_grp, stateUpdates):
        return self._handler.sendMetricStateUpdates(mdib_version_grp, stateUpdates)

    def sendAlertStateUpdates(self, mdib_version_grp, stateUpdates):
        return self._handler.sendAlertStateUpdates(mdib_version_grp, stateUpdates)

    def sendComponentStateUpdates(self, mdib_version_grp, stateUpdates):
        return self._handler.sendComponentStateUpdates(mdib_version_grp, stateUpdates)

    def sendContextStateUpdates(self, mdib_version_grp, stateUpdates):
        return self._handler.sendContextStateUpdates(mdib_version_grp, stateUpdates)

    def sendOperationalStateUpdates(self, mdib_version_grp, stateUpdates):
        return self._handler.sendOperationalStateUpdates(mdib_version_grp, stateUpdates)

    def sendRealtimeSamplesStateUpdates(self, mdib_version_grp, stateUpdates):
        return self._handler.sendRealtimeSamplesStateUpdates(mdib_version_grp, stateUpdates)

    def sendDescriptorUpdates(self, mdib_version_grp, updated, created, deleted, updated_states):
        return self._handler.sendDescriptorUpdates(mdib_version_grp, updated, created, deleted, updated_states)

    def setUsedCompression(self, *compression_methods):
        return self._handler.setUsedCompression(*compression_methods)

    @property
    def product_roles(self):
        return self._handler.product_roles

    @product_roles.setter
    def product_roles(self, product_roles):
        self._handler.product_roles = product_roles



import ifaddr

_IP_BLACKLIST = ('0.0.0.0', None)  # None can happen if an adapter does not have any IP address


def get_ipv4_ips():
    """returns a list of ifaddr._shared.IP objects (ip, nice_name, network_prefix)"""
    ips = []
    all_adapters = ifaddr.get_adapters()
    for adapter in all_adapters:
        for adapter.ip in adapter.ips:
            if adapter.ip.is_IPv4 and adapter.ip.ip not in _IP_BLACKLIST:
                ips.append(adapter.ip)
    return ips


def get_ipv4_addresses():
    """returns ip addresses"""
    return [ip.ip for ip in get_ipv4_ips()]


def get_ip_for_adapter(adapter_name):
    """returns a single ip address or None"""
    ip_objects = get_ipv4_ips()
    for ip_object in ip_objects:
        if ip_object.nice_name == adapter_name:
            return ip_object.ip
    return None



import threading
from .observables import bind, unbind


class Error(Exception):
    """Base class for all ValuesCollector-related exceptions."""
    pass

class CancelledError(Error):
    """The ValuesCollector was cancelled."""
    pass

class TimeoutError(Error):
    """The operation exceeded the given deadline."""
    pass



class SingleValueCollector(object):
    """ collects next data item from an observable.
    usage:
    assuming myObj has an ObservableProperty named 'myProperty':
    collector = SingleValueCollector(myObj, 'myProperty') # collector will now retrieve and store the value of the next set value of observable property
    result = collector.result(timeout=0.1) # wait until result is available (or timeout
        some other thread:     myObj.myProperty = 42
    => now call of  collector.result() returns, result = 42
    """
    # Possible states
    PENDING = 'PENDING'
    FINISHED = 'FINISHED'
    CLOSED = 'CLOSED'
    
    def __init__(self, obj, propName):
        self._obj = obj
        self._propName = propName
        self._cond = threading.Condition()
        bind(obj, **{propName:self._onData})
        self._state = self.PENDING
        self._exception = None
        self._result = None
        
    def _onData(self, data):
        if self._state == self.CLOSED:
            return
        with self._cond:
            self._result = data
            self._state = self.FINISHED
            unbind(self._obj, **{self._propName:self._onData})
            self._cond.notify_all()

    def __get_result(self):
        if self._exception:
            raise # re-raise
        else:
            return self._result

    def result(self, timeout=None):
        if self._state == self.CLOSED:
            raise RuntimeError('SingleValueCollector is already closed')
        with self._cond:
            if self._state == self.FINISHED:
                self._state = self.CLOSED
                return self.__get_result()

            self._cond.wait(timeout)

            if self._state == self.FINISHED:
                self._state = self.CLOSED
                return self.__get_result()
            else:
                unbind(self._obj, **{self._propName:self._onData})
                self._state = self.CLOSED
                raise TimeoutError()

    def restart(self):
        """ Start to capture another value
        """
        if self._state != self.CLOSED:
            raise RuntimeError('SingleValueCollector is still active')
        bind(self._obj, **{self._propName:self._onData})
        self._state = self.PENDING
        self._exception = None
        self._result = None



class ValuesCollector(SingleValueCollector):
    """ collects multiple data from an observable.
    usage:
    assuming myObj has an ObservableProperty named 'myProperty':
    collector = ValuesCollector(myObj, 'myProperty', 2) # collector will now retrieve and store the value of the next 2 set value of observable property
    result = collector.result(timeout=0.1) # wait until result is available (or timeout
        some other thread:     myObj.myProperty = 42
                               myObj.myProperty = 43
    => now call of  collector.result() returns, result = [42, 43]
    """
    def __init__(self, obj, propName, n):
        super(ValuesCollector, self).__init__(obj, propName)
        self._n = n
        self._result = []


    def _onData(self, data):
        if self._state == self.CLOSED:
            return
        with self._cond:
            if self._state == self.FINISHED:
                return
            self._result.append(data)
            if len(self._result) >= self._n:
                self._state = self.FINISHED
                unbind(self._obj, **{self._propName:self._onData})
                self._cond.notify_all()




import logging
import os
import threading
import time
import traceback
import typing
import urllib
import uuid
from collections import namedtuple

from lxml import etree as etree_

import sdc11073.certloader
from . import httpserver
from . import intervaltimer
from . import sco
from . import subscriptionmgr
from .localizationservice import LocalizationService
from .sdcservicesimpl import GetService, SetService, StateEventService, ContainmentTreeService, ContextService, \
    WaveformService, DescriptionEventService
from .sdcservicesimpl import SOAPActionDispatcher, DPWSHostedService
from .. import compression
from .. import loghelper
from .. import namespaces
from .. import pmtypes
from .. import pysoap
from .. import wsdiscovery
from ..location import SdcLocation

Soap12Envelope = pysoap.soapenvelope.Soap12Envelope

Prefix = namespaces.Prefix_Namespace

# default ssl context data
here = os.path.dirname(__file__)
caFolder = os.path.join(os.path.dirname(here), 'ca')
_ssl_certfile = os.path.join(caFolder,
                             'sdccert.pem')  # this is the certification chain ( contains root ca and signed public key
_ssl_keyfile = os.path.join(caFolder, 'userkey.pem')  # this is the private key of own certificate
_ssl_cacert = os.path.join(caFolder, 'cacert.pem')  # this is the common root ca that signed all sdc devices
_ssl_passwd = 'dummypass'  # 'Phase1' #dummypass
_ssl_cypherfile = os.path.join(caFolder, 'cyphers.json')  # Json file that determines ciphers to be used

PeriodicStates = namedtuple('PeriodicStates', 'mdib_version states')


class SdcHandler_Base(object):
    """ This is the base class for the sdc device handler. It contains all functionality of a device except the definition of the hosted services.
    These must be instantiated in a derived class."""

    SSL_CIPHERS = 'HIGH:!3DES:!DSS:!aNULL@STRENGTH'

    WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE = 0.2  # warn limit when real time samples cannot be sent in time (typically because receiver is too slow)
    WARN_RATE_REALTIMESAMPLES_BEHIND_SCHEDULE = 5  # max. every x seconds a message

    DEFAULT_CONTEXTSTATES_IN_GETMDIB = True  # defines if getMdib and getMdStates contain context states or not.
    # This is a default, it can be overidden per instande in
    # member "contextstates_in_getmdib".
    defaultInstanceIdentifiers = (pmtypes.InstanceIdentifier(root='rootWithNoMeaning', extensionString='System'),)

    _ssl_context_container: typing.Optional[sdc11073.certloader.SSLContextContainer]

    def __init__(self, my_uuid, ws_discovery, model, device, deviceMdibContainer, validate=True,
                 roleProvider=None, sslContext=None,
                 logLevel=None, max_subscription_duration=7200, log_prefix='', chunked_messages=False,
                 ssl_context_container: sdc11073.certloader.SSLContextContainer = None):  # pylint:disable=too-many-arguments
        """
        :param uuid: a string that becomes part of the devices url (no spaces, no special characters please. This could cause an invalid url!).
                     Parameter can be None, in this case a random uuid string is generated.
        :param ws_discovery: reference to the wsDiscovery instance
        :param model: a pysoap.soapenvelope.DPWSThisModel instance
        :param device: a pysoap.soapenvelope.DPWSThisDevice instance
        :param deviceMdibContainer: a DeviceMdibContainer instance
        :param roleProvider: handles the operation calls
        :param sslContext: if not None, this context is used and https url is used. Otherwise http
        :param logLevel: if not None, the "sdc.device" logger will use this level
        :param max_subscription_duration: max. possible duration of a subscription, default is 7200 seconds
        :param ident: names a device, used for logging
        :param ssl_context_container: container containing a client and a server context
        """
        self._my_uuid = my_uuid or uuid.uuid4()
        self._wsdiscovery = ws_discovery
        self.model = model
        self.device = device
        self._mdib = deviceMdibContainer
        self._log_prefix = log_prefix
        self._mdib.log_prefix = log_prefix
        self._validate = validate

        if sslContext:
            self._ssl_context_container = sdc11073.certloader.SSLContextContainer(client_context=sslContext,
                                                                                  server_context=sslContext)
        else:
            self._ssl_context_container = ssl_context_container

        self._compression_methods = compression.encodings[:]
        self._httpServerThread = None
        self._setupLogging(logLevel)
        self._logger = loghelper.getLoggerAdapter('sdc.device', log_prefix)

        self.chunked_messages = chunked_messages
        self.contextstates_in_getmdib = self.DEFAULT_CONTEXTSTATES_IN_GETMDIB  # can be overridden per instance
        # hostDispatcher provides data of the sdc device itself
        self._hostDispatcher = self._mkHostDispatcher()

        self._GetDispatcher = None
        self._LocalizationDispatcher = None
        self._GetServiceHosted = None
        self._ContextDispatcher = None
        self._DescriptionEventDispatcher = None
        self._StateEventDispatcher = None
        self._WaveformDispatcher = None
        self._SdcServiceHosted = None
        self.__SetDispatcher = None
        self._SetServiceHosted = None
        self._hostedServices = []
        self._url_dispatcher = None
        self._rtSampleSendThread = None
        self._runRtSampleThread = False
        self.collectRtSamplesPeriod = 0.1  # in seconds
        if self._ssl_context_container is not None:
            self._urlschema = 'https'
        else:
            self._urlschema = 'http'

        self.dpwsHost = None
        self._subscriptionsManager = self._mkSubscriptionManager(max_subscription_duration)
        self._scoOperationsRegistry = self._mkScoOperationsRegistry(handle='_sco')

        deviceMdibContainer.setSdcDevice(self)

        self.product_roles = roleProvider
        if self.product_roles is None:
            self.mkDefaultRoleHandlers()

        self._location = None
        self._periodic_reports_lock = threading.Lock()
        self._periodic_reports_thread = None
        self._run_periodic_reports_thread = None
        self._periodic_reports_interval = None
        self._periodic_metric_reports = []
        self._periodic_alert_reports = []
        self._periodic_component_state_reports = []
        self._periodic_context_state_reports = []
        self._periodic_operational_state_reports = []

        self.xml_validator = self._mdib.sdc_definitions.xml_validator

    def mkScopes(self):
        scopes = []
        locations = self._mdib.contextStates.NODETYPE.get(namespaces.domTag('LocationContextState'), [])
        assoc_loc = [l for l in locations if l.ContextAssociation == pmtypes.ContextAssociation.ASSOCIATED]
        for loc in assoc_loc:
            dr_loc = SdcLocation(fac=loc.Facility, poc=loc.PoC, bed=loc.Bed, bld=loc.Building,
                                 flr=loc.Floor, rm=loc.Room)
            scopes.append(wsdiscovery.Scope(dr_loc.scopeStringSdc))

        for nodetype, scheme in (
                ('OperatorContextDescriptor', 'sdc.ctxt.opr'),
                ('EnsembleContextDescriptor', 'sdc.ctxt.ens'),
                ('WorkflowContextDescriptor', 'sdc.ctxt.wfl'),
                ('MeansContextDescriptor', 'sdc.ctxt.mns'),
        ):
            descriptors = self._mdib.descriptions.NODETYPE.get(namespaces.domTag(nodetype), [])
            for descriptor in descriptors:
                states = self._mdib.contextStates.descriptorHandle.get(descriptor.Handle, [])
                assoc_st = [s for s in states if s.ContextAssociation == pmtypes.ContextAssociation.ASSOCIATED]
                for st in assoc_st:
                    for ident in st.Identification:
                        scopes.append(wsdiscovery.Scope('{}:/{}/{}'.format(scheme, urllib.parse.quote_plus(ident.Root),
                                                                           urllib.parse.quote_plus(ident.Extension))))

        scopes.extend(self._getDeviceComponentBasedScopes())
        scopes.append(wsdiscovery.Scope('sdc.mds.pkp:1.2.840.10004.20701.1.1'))  # key purpose Service provider
        return scopes

    def _getDeviceComponentBasedScopes(self):
        """
        SDC: For every instance derived from pm:AbstractComplexDeviceComponentDescriptor in the MDIB an
        SDC SERVICE PROVIDER SHOULD include a URIencoded pm:AbstractComplexDeviceComponentDescriptor/pm:Type
        as dpws:Scope of the MDPWS discovery messages. The URI encoding conforms to the given Extended Backus-Naur Form.
        E.G.  sdc.cdc.type:///69650, sdc.cdc.type:/urn:oid:1.3.6.1.4.1.3592.2.1.1.0//DN_VMD
        After discussion with David: use only MDSDescriptor, VmdDescriptor makes no sense.
        :return: a set of scopes
        """
        scopes = set()
        for t in (namespaces.domTag('MdsDescriptor'),):
            descriptors = self._mdib.descriptions.NODETYPE.get(t)
            for d in descriptors:
                if d.Type is not None:
                    cs = '' if d.Type.CodingSystem == pmtypes.DefaultCodingSystem else d.Type.CodingSystem
                    csv = d.Type.CodingSystemVersion or ''
                    sc = wsdiscovery.Scope('sdc.cdc.type:/{}/{}/{}'.format(cs, csv, d.Type.Code))
                    scopes.add(sc)
        return scopes

    def _mkHostDispatcher(self):
        hostDispatcher = SOAPActionDispatcher()
        hostDispatcher.register_soapActionCallback('{}/Get'.format(Prefix.WXF.namespace), self._onGetMetaData)
        hostDispatcher.register_soapActionCallback('{}/Probe'.format(Prefix.WSD.namespace), self._onProbeRequest)
        hostDispatcher.epr = '/' + str(self._my_uuid.hex)
        return hostDispatcher

    def _mkSubscriptionManager(self, max_subscription_duration):
        return subscriptionmgr.SubscriptionsManager(self._ssl_context_container,
                                                    self._mdib.sdc_definitions,
                                                    self._compression_methods,
                                                    max_subscription_duration,
                                                    log_prefix=self._log_prefix,
                                                    chunked_messages=self.chunked_messages)

    def _mkScoOperationsRegistry(self, handle):
        return sco.ScoOperationsRegistry(self._subscriptionsManager, self._mdib, handle, log_prefix=self._log_prefix)

    def mkDefaultRoleHandlers(self):
        from .. import roles
        self.product_roles = roles.product.MinimalProduct(self._log_prefix)

    # @property
    # def _bmmSchema(self):
    #     return None if not self._validate else self._mdib.bicepsSchema.bmmSchema

    @property
    def shallValidate(self):
        return self._validate

    @property
    def mdib(self):
        return self._mdib

    @property
    def subscriptionsManager(self):
        return self._subscriptionsManager

    @property
    def epr(self):
        # End Point Reference, e.g 'urn:uuid:8c26f673-fdbf-4380-b5ad-9e2454a65b6b'
        return self._my_uuid.urn

    @property
    def path_prefix(self):
        # http path prefix of service e.g '8c26f673-fdbf-4380-b5ad-9e2454a65b6b'
        return self._my_uuid.hex

    def registerOperation(self, operation):
        self._scoOperationsRegistry.registerOperation(operation)

    def unRegisterOperationByHandle(self, operationHandle):
        self._scoOperationsRegistry.registerOperation(operationHandle)

    def getOperationByHandle(self, operationHandle):
        return self._scoOperationsRegistry.getOperationByHandle(operationHandle)

    def enqueueOperation(self, operation, request):
        return self._scoOperationsRegistry.enqueueOperation(operation, request)

    def dispatchGetRequest(self, parseResult, headers):
        """ device itself can also handle GET requests. This is the handler"""
        return self._hostDispatcher.dispatchGetRequest(parseResult, headers)

    def _startServices(self, shared_http_server=None, http_server_timeout=15.0):
        """ start the services"""
        self._logger.info('starting services, addr = {}', self._wsdiscovery.getActiveAddresses())

        self._scoOperationsRegistry.startWorker()
        if shared_http_server:
            self._httpServerThread = shared_http_server
        else:
            self._httpServerThread = httpserver.HttpServerThread(
                my_ipaddress='0.0.0.0',
                sslContext=self._ssl_context_container.server_context if self._ssl_context_container else None,
                supportedEncodings=self._compression_methods,
                log_prefix=self._log_prefix,
                chunked_responses=self.chunked_messages)

            # first start http server, the services need to know the ip port number
            self._httpServerThread.start()
            event_is_set = self._httpServerThread.started_evt.wait(timeout=http_server_timeout)
            if not event_is_set:
                self._logger.error('Cannot start device, start event of http server not set.')
                raise RuntimeError('Cannot start device, start event of http server not set.')

        host_ips = self._wsdiscovery.getActiveAddresses()
        self._url_dispatcher = httpserver.HostedServiceDispatcher(self._mdib.sdc_definitions, self._logger)
        self._httpServerThread.devices_dispatcher.register_device_dispatcher(self.path_prefix, self._url_dispatcher)
        if len(host_ips) == 0:
            self._logger.error('Cannot start device, there is no IP address to bind it to.')
            raise RuntimeError('Cannot start device, there is no IP address to bind it to.')

        port = self._httpServerThread.my_port
        if port is None:
            self._logger.error('Cannot start device, could not bind HTTP server to a port.')
            raise RuntimeError('Cannot start device, could not bind HTTP server to a port.')

        base_urls = []  # e.g https://192.168.1.5:8888/8c26f673-fdbf-4380-b5ad-9e2454a65b6b; list has one member for each used ip address
        for addr in host_ips:
            base_urls.append(
                urllib.parse.SplitResult(self._urlschema, '{}:{}'.format(addr, port), self.path_prefix, query=None,
                                         fragment=None))
        self.dpwsHost = pysoap.soapenvelope.DPWSHost(
            endpointReferencesList=[pysoap.soapenvelope.WsaEndpointReferenceType(self.epr)],
            typesList=self._mdib.sdc_definitions.MedicalDeviceTypesFilter)
        # register two addresses for hostDispatcher: '' and /<uuid>
        self._url_dispatcher.register_hosted_service(self._hostDispatcher)

        self._register_hosted_services(base_urls)

        for host_ip in host_ips:
            self._logger.info('serving Services on {}:{}', host_ip, port)
        self._subscriptionsManager.setBaseUrls(base_urls)

    def startAll(self, startRealtimeSampleLoop=True, periodic_reports_interval=None, shared_http_server=None,
                 http_server_timeout=15.0):
        if self.product_roles is not None:
            self.product_roles.initOperations(self._mdib, self._scoOperationsRegistry)

        self._startServices(shared_http_server, http_server_timeout)
        if startRealtimeSampleLoop:
            self._runRtSampleThread = True
            self._rtSampleSendThread = threading.Thread(target=self._rt_sample_sendloop, name='DevRtSampleSendLoop')
            self._rtSampleSendThread.daemon = True
            self._rtSampleSendThread.start()
        if periodic_reports_interval:
            self._run_periodic_reports_thread = True
            self._periodic_reports_interval = periodic_reports_interval
            self._periodic_reports_thread = threading.Thread(target=self._periodic_reports_send_loop,
                                                             name='DevPeriodicSendLoop')
            self._periodic_reports_thread.daemon = True
            self._periodic_reports_thread.start()

    def stopAll(self, closeAllConnections, sendSubscriptionEnd):
        if self._rtSampleSendThread is not None:
            self._runRtSampleThread = False
            self._rtSampleSendThread.join()
            self._rtSampleSendThread = None
            if self._run_periodic_reports_thread:
                self._run_periodic_reports_thread = False
                self._periodic_reports_thread.join()

        self._subscriptionsManager.endAllSubscriptions(sendSubscriptionEnd)
        self._scoOperationsRegistry.stopWorker()
        self._httpServerThread.stop(closeAllConnections)
        try:
            self._wsdiscovery.clearService(self.epr)
        except KeyError:
            self._logger.info('epr "{}" not known in self._wsdiscovery'.format(self.epr))

        if self.product_roles is not None:
            self.product_roles.stop()

    def getXAddrs(self):
        addresses = self._wsdiscovery.getActiveAddresses()  # these own IP addresses are currently used by discovery
        port = self._httpServerThread.my_port
        xaddrs = []
        for xa in addresses:
            xaddrs.append('{}://{}:{}/{}'.format(self._urlschema, xa, port, self.path_prefix))
        return xaddrs

    def _onGetMetaData(self, httpHeader, request):
        self._logger.info('_onGetMetaData')
        _nsm = self._mdib.nsmapper
        response = pysoap.soapenvelope.Soap12Envelope(_nsm.docNssmap)
        replyAddress = request.address.mkReplyAddress('{}/GetResponse'.format(Prefix.WXF.namespace))
        replyAddress.to = namespaces.WSA_ANONYMOUS
        replyAddress.messageId = uuid.uuid4().urn
        response.addHeaderObject(replyAddress)
        metaDataNode = self._mkMetaDataNode()
        response.addBodyElement(metaDataNode)
        self._logger.debug('returned meta data = {}', response.as_xml(pretty=False))
        return response

    def _onProbeRequest(self, httpHeader, request):
        _nsm = namespaces.DocNamespaceHelper()
        response = pysoap.soapenvelope.Soap12Envelope(_nsm.docNssmap)
        replyAddress = request.address.mkReplyAddress('{}/ProbeMatches'.format(Prefix.WSD.namespace))
        replyAddress.to = namespaces.WSA_ANONYMOUS
        replyAddress.messageId = uuid.uuid4().urn
        response.addHeaderObject(replyAddress)
        probe_match_node = etree_.Element(namespaces.wsdTag('Probematch'),
                                          nsmap=_nsm.docNssmap)
        types = etree_.SubElement(probe_match_node, namespaces.wsdTag('Types'))
        types.text = '{}:Device {}:MedicalDevice'.format(Prefix.DPWS.prefix, Prefix.MDPWS.prefix)
        scopes = etree_.SubElement(probe_match_node, namespaces.wsdTag('Scopes'))
        scopes.text = ''
        xaddrs = etree_.SubElement(probe_match_node, namespaces.wsdTag('XAddrs'))
        xaddrs.text = ' '.join(self.getXAddrs())
        response.addBodyElement(probe_match_node)
        return response

    def _validateDPWS(self, node):
        if not self.shallValidate:
            return
        try:
            self.xml_validator.assertValid(node)
        except etree_.DocumentInvalid as ex:
            tmp_str = etree_.tostring(node, pretty_print=True).decode('utf-8')
            self._logger.error('invalid dpws: {}\ndata = {}', ex, tmp_str)
            raise

    def _mkMetaDataNode(self):
        metaDataNode = etree_.Element(namespaces.wsxTag('Metadata'),
                                      nsmap=self._mdib.nsmapper.docNssmap)

        # ThisModel
        metaDataSectionNode = etree_.SubElement(metaDataNode,
                                                namespaces.wsxTag('MetadataSection'),
                                                attrib={'Dialect': '{}/ThisModel'.format(namespaces.nsmap['dpws'])})
        self.model.asEtreeSubNode(metaDataSectionNode)
        self._validateDPWS(metaDataSectionNode[-1])

        # ThisDevice
        metaDataSectionNode = etree_.SubElement(metaDataNode,
                                                namespaces.wsxTag('MetadataSection'),
                                                attrib={'Dialect': '{}/ThisDevice'.format(namespaces.nsmap['dpws'])})
        self.device.asEtreeSubNode(metaDataSectionNode)

        self._validateDPWS(metaDataSectionNode[-1])

        # Relationship
        metaDataSectionNode = etree_.SubElement(metaDataNode,
                                                namespaces.wsxTag('MetadataSection'),
                                                attrib={'Dialect': '{}/Relationship'.format(namespaces.nsmap['dpws'])})
        relationshipNode = etree_.SubElement(metaDataSectionNode,
                                             namespaces.dpwsTag('Relationship'),
                                             attrib={'Type': '{}/host'.format(namespaces.nsmap['dpws'])})

        self.dpwsHost.asEtreeSubNode(relationshipNode)
        self._validateDPWS(relationshipNode[-1])

        # add all hosted services:
        for service in self._hostedServices:
            service.hostedInf.asEtreeSubNode(relationshipNode)
            self._validateDPWS(relationshipNode[-1])
        return metaDataNode

    def _store_for_periodic_report(self, mdib_version, state_updates, dest_list):
        if self._run_periodic_reports_thread:
            copied_updates = [s.mkCopy() for s in state_updates]
            with self._periodic_reports_lock:
                dest_list.append(PeriodicStates(mdib_version, copied_updates))

    def sendMetricStateUpdates(self, mdib_version_grp, stateUpdates):
        self._logger.debug('sending metric state updates {}', stateUpdates)
        self._subscriptionsManager.sendEpisodicMetricReport(stateUpdates, self._mdib.nsmapper, mdib_version_grp)
        self._store_for_periodic_report(mdib_version_grp.mdib_version, stateUpdates, self._periodic_metric_reports)

    def sendAlertStateUpdates(self, mdib_version_grp, stateUpdates):
        self._logger.debug('sending alert updates {}', stateUpdates)
        self._subscriptionsManager.sendEpisodicAlertReport(stateUpdates, self._mdib.nsmapper, mdib_version_grp)
        self._store_for_periodic_report(mdib_version_grp.mdib_version, stateUpdates, self._periodic_alert_reports)

    def sendComponentStateUpdates(self, mdib_version_grp, stateUpdates):
        self._logger.debug('sending component state updates {}', stateUpdates)
        self._subscriptionsManager.sendEpisodicComponentStateReport(stateUpdates, self._mdib.nsmapper, mdib_version_grp)
        self._store_for_periodic_report(mdib_version_grp.mdib_version, stateUpdates,
                                        self._periodic_component_state_reports)

    def sendContextStateUpdates(self, mdib_version_grp, stateUpdates):
        self._logger.debug('sending context updates {}', stateUpdates)
        self._subscriptionsManager.sendEpisodicContextReport(stateUpdates, self._mdib.nsmapper, mdib_version_grp)
        self._store_for_periodic_report(mdib_version_grp.mdib_version, stateUpdates,
                                        self._periodic_context_state_reports)

    def sendOperationalStateUpdates(self, mdib_version_grp, stateUpdates):
        self._logger.debug('sending operational state updates {}', stateUpdates)
        self._subscriptionsManager.sendEpisodicOperationalStateReport(stateUpdates, self._mdib.nsmapper,
                                                                      mdib_version_grp)
        self._store_for_periodic_report(mdib_version_grp.mdib_version, stateUpdates,
                                        self._periodic_operational_state_reports)

    def sendRealtimeSamplesStateUpdates(self, mdib_version_grp, stateUpdates):
        self._logger.debug('sending real time sample state updates {}', stateUpdates)
        self._subscriptionsManager.sendRealtimeSamplesReport(stateUpdates, self._mdib.nsmapper, mdib_version_grp)

    def sendDescriptorUpdates(self, mdib_version_grp, updated, created, deleted, updated_states):
        self._logger.debug('sending descriptor updates updated={} created={} deleted={}', updated, created, deleted)
        self._subscriptionsManager.sendDescriptorUpdates(updated, created, deleted, updated_states,
                                                         self._mdib.nsmapper, mdib_version_grp)

    def _rt_sample_sendloop(self):
        """Periodically send waveform samples."""
        # start delayed in order to have a fully initialized device when waveforms start
        # (otherwise timing issues might happen)
        time.sleep(0.1)
        timer = intervaltimer.IntervalTimer(periodInSeconds=self.collectRtSamplesPeriod)
        while self._runRtSampleThread:
            behindScheduleSeconds = timer.waitForNextIntervalBegin()
            try:
                self._mdib.update_all_rt_samples()  # update from waveform generators
                self._logWaveformTiming(behindScheduleSeconds)
            except Exception:
                self._logger.warn(' could not update real time samples: {}', traceback.format_exc())

    def _periodic_reports_send_loop(self):
        """This is a very basic implementation of periodic reports, it only supports fixed interval.
        It does not care about retrievability settings in the mdib.
        """
        self._logger.debug('_periodic_reports_send_loop start')
        time.sleep(0.1)  # start delayed
        timer = intervaltimer.IntervalTimer(periodInSeconds=self._periodic_reports_interval)
        while self._run_periodic_reports_thread:
            timer.waitForNextIntervalBegin()
            self._logger.debug('_periodic_reports_send_loop')
            for reports_list, send_func, msg in [
                (self._periodic_metric_reports, self._subscriptionsManager.sendPeriodicMetricReport, 'metric'),
                (self._periodic_alert_reports, self._subscriptionsManager.sendPeriodicAlertReport, 'alert'),
                (self._periodic_component_state_reports, self._subscriptionsManager.sendPeriodicComponentStateReport,
                 'component'),
                (self._periodic_context_state_reports, self._subscriptionsManager.sendPeriodicContextReport, 'context'),
                (
                        self._periodic_operational_state_reports,
                        self._subscriptionsManager.sendPeriodicOperationalStateReport,
                        'operational'),
            ]:
                tmp = None
                with self._periodic_reports_lock:
                    if reports_list:
                        tmp = reports_list[:]
                        del reports_list[:]
                if tmp:
                    self._logger.debug('send periodic %s report', msg)
                    send_func(tmp, self._mdib.nsmapper, self.mdib.mdib_version_group)

    def _setupLogging(self, logLevel):
        loghelper.ensureLogStream()
        if logLevel is None:
            return
        deviceLog = logging.getLogger('sdc.device')
        deviceLog.setLevel(logLevel)

    def _logWaveformTiming(self, behindScheduleSeconds):
        try:
            lastLogTime = self._lastLogTime
        except AttributeError:
            self._lastLogTime = 0
            lastLogTime = self._lastLogTime
        try:
            lastLoggedDelay = self._lastLoggedDelay
        except AttributeError:
            self._lastLoggedDelay = 0
            lastLoggedDelay = self._lastLoggedDelay

        # max. one log per second
        now = time.monotonic()
        if now - lastLogTime < self.WARN_RATE_REALTIMESAMPLES_BEHIND_SCHEDULE:
            return
        if lastLoggedDelay >= self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE and behindScheduleSeconds < self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE:
            self._logger.info('RealTimeSampleTimer delay is back inside limit of {:.2f} seconds (mdib version={}',
                              self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE, self._mdib.mdibVersion)
            self._lastLoggedDelay = behindScheduleSeconds
            self._lastLogTime = now
        elif behindScheduleSeconds >= self.WARN_LIMIT_REALTIMESAMPLES_BEHIND_SCHEDULE:
            self._logger.warn('RealTimeSampleTimer is {:.4f} seconds behind schedule (mdib version={})',
                              behindScheduleSeconds, self._mdib.mdibVersion)
            self._lastLoggedDelay = behindScheduleSeconds
            self._lastLogTime = now

    def setUsedCompression(self, *compression_methods):
        # update list in place
        del self._compression_methods[:]
        self._compression_methods.extend(compression_methods)


class SdcHandler_Full(SdcHandler_Base):
    """ This class instantiates all port types."""

    def _register_hosted_services(self, base_urls):
        # register all services with their endpoint references acc. to sdc standard
        actions = self._mdib.sdc_definitions.Actions

        self._GetDispatcher = GetService('GetService', self)
        self._LocalizationDispatcher = LocalizationService('LocalizationService', self)
        offeredSubscriptions = []
        self._GetServiceHosted = DPWSHostedService(self, base_urls, 'Get',
                                                   [self._GetDispatcher, self._LocalizationDispatcher],
                                                   offeredSubscriptions)
        self._url_dispatcher.register_hosted_service(self._GetServiceHosted)

        # grouped acc to sdc REQ 0035
        self._ContextDispatcher = ContextService('ContextService', self)
        self._DescriptionEventDispatcher = DescriptionEventService('DescriptionEventService', self)
        self._StateEventDispatcher = StateEventService('StateEventService', self)
        self._WaveformDispatcher = WaveformService('WaveformService', self)

        offeredSubscriptions = [actions.EpisodicContextReport,
                                actions.DescriptionModificationReport,
                                actions.EpisodicMetricReport,
                                actions.EpisodicAlertReport,
                                actions.EpisodicComponentReport,
                                actions.EpisodicOperationalStateReport,
                                actions.Waveform,
                                actions.SystemErrorReport,
                                actions.PeriodicMetricReport,
                                actions.PeriodicAlertReport,
                                actions.PeriodicContextReport,
                                actions.PeriodicComponentReport,
                                actions.PeriodicOperationalStateReport
                                ]

        self._SdcServiceHosted = DPWSHostedService(self, base_urls, 'StateEvent',
                                                   [self._ContextDispatcher,
                                                    self._DescriptionEventDispatcher,
                                                    self._StateEventDispatcher,
                                                    self._WaveformDispatcher],
                                                   offeredSubscriptions)
        self._url_dispatcher.register_hosted_service(self._SdcServiceHosted)

        self.__SetDispatcher = SetService('SetService', self)
        offeredSubscriptions = [actions.OperationInvokedReport]

        self._SetServiceHosted = DPWSHostedService(self, base_urls, 'Set', [self.__SetDispatcher], offeredSubscriptions)
        self._url_dispatcher.register_hosted_service(self._SetServiceHosted)

        self._ContainmentTreeDispatcher = ContainmentTreeService('ContainmentTreeService', self)
        offeredSubscriptions = []
        self._ContainmentTreeServiceHosted = DPWSHostedService(self, base_urls, 'ContainmentTree',
                                                               [self._ContainmentTreeDispatcher], offeredSubscriptions)
        self._url_dispatcher.register_hosted_service(self._ContainmentTreeServiceHosted)
        self._hostedServices = [self._GetServiceHosted,
                                self._SdcServiceHosted,
                                self._SetServiceHosted,
                                self._ContainmentTreeServiceHosted]


class SdcHandler_Minimal(SdcHandler_Base):
    """This class instantiates only GetService and LocalizationService"""

    def _register_hosted_services(self, base_urls):
        self._GetDispatcher = GetService('GetService', self)
        self._LocalizationDispatcher = LocalizationService('LocalizationService', self)
        offeredSubscriptions = []
        self._GetServiceHosted = DPWSHostedService(self, base_urls, 'Get',
                                                   [self._GetDispatcher, self._LocalizationDispatcher],
                                                   offeredSubscriptions)
        self._url_dispatcher.register_hosted_service(self._GetServiceHosted)

        self._url_dispatcher.register_hosted_service(self._ContainmentTreeServiceHosted)
        self._hostedServices = [self._GetServiceHosted]
